# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import jax
from jax import numpy as jnp


@jax.custom_jvp
def doubly_truncated_power_law_log_norm_constant(alpha, low, high):
    neq_neg1_mask = jnp.not_equal(alpha, -1.0)
    neq_neg1_alpha = jnp.where(neq_neg1_mask, alpha, 0.0)

    def neq_neg1_fn():
        one_more_alpha = 1.0 + neq_neg1_alpha
        return jnp.log(
            (jnp.power(high, one_more_alpha) - jnp.power(low, one_more_alpha))
            / (one_more_alpha)
        )

    def eq_neg1_fn():
        return jnp.log(jnp.log(high) - jnp.log(low))

    return jnp.where(neq_neg1_mask, neq_neg1_fn(), eq_neg1_fn())


@doubly_truncated_power_law_log_norm_constant.defjvp
def doubly_truncated_power_law_log_norm_constant_jvp(primals, tangents):
    alpha, low, high = primals
    alpha_t, low_t, high_t = tangents

    primal_out = doubly_truncated_power_law_log_norm_constant(*primals)

    log_low = jnp.log(low)
    log_high = jnp.log(high)

    # Mask and alpha values
    delta_eq_neg1 = 1e-5
    neq_neg1_mask = jnp.not_equal(alpha, -1.0)
    neq_neg1_alpha = jnp.where(neq_neg1_mask, alpha, 0.0)

    # Alpha tangent with approximation
    # Variable part for all values alpha unequal -1
    def alpha_tangent_variable(alpha):
        one_more_alpha = 1.0 + alpha
        low_pow_one_more_alpha = jnp.power(low, one_more_alpha)
        high_pow_one_more_alpha = jnp.power(high, one_more_alpha)
        return -jnp.reciprocal(one_more_alpha) + (
            high_pow_one_more_alpha * log_high - low_pow_one_more_alpha * log_low
        ) / (high_pow_one_more_alpha - low_pow_one_more_alpha)

    # Alpha tangent
    alpha_tangent = jnp.where(
        neq_neg1_mask,
        alpha_tangent_variable(neq_neg1_alpha),
        # Approximate derivate with right an lefthand approximation
        (
            alpha_tangent_variable(alpha - delta_eq_neg1)
            + alpha_tangent_variable(alpha + delta_eq_neg1)
        )
        * 0.5,
    )

    # High and low tangents for alpha unequal -1
    one_more_alpha = 1.0 + neq_neg1_alpha
    low_pow_one_more_alpha = jnp.power(low, one_more_alpha)
    high_pow_one_more_alpha = jnp.power(high, one_more_alpha)
    change_neq_neg1_alpha = high_pow_one_more_alpha - low_pow_one_more_alpha
    low_tangent_neq_neg1 = (
        -one_more_alpha * jnp.power(low, neq_neg1_alpha) / change_neq_neg1_alpha
    )
    high_tangent_neq_neg1 = (
        one_more_alpha * jnp.power(high, neq_neg1_alpha) / change_neq_neg1_alpha
    )

    # High and low tangents for alpha equal -1
    change_eq_neg1_alpha = log_high - log_low
    low_tangent_eq_neg1 = -jnp.reciprocal(low * change_eq_neg1_alpha)
    high_tangent_eq_neg1 = jnp.reciprocal(high * change_eq_neg1_alpha)

    # High and low tangents
    low_tangent = jnp.where(neq_neg1_mask, low_tangent_neq_neg1, low_tangent_eq_neg1)
    high_tangent = jnp.where(neq_neg1_mask, high_tangent_neq_neg1, high_tangent_eq_neg1)

    # Final tangents
    tangent_out = alpha_tangent * alpha_t + low_tangent * low_t + high_tangent * high_t
    return primal_out, tangent_out


def doubly_truncated_power_law_log_prob(x, alpha, low, high):
    return alpha * jnp.log(x) - doubly_truncated_power_law_log_norm_constant(
        alpha, low, high
    )


@jax.custom_jvp
def doubly_truncated_power_law_cdf(x, alpha, low, high):
    # source https://github.com/pyro-ppl/numpyro/blob/94f4b99710d855bea456210cf91e6e55eeac3926/numpyro/distributions/truncated.py#L545-L565
    neq_neg1_mask = jnp.not_equal(alpha, -1.0)
    neq_neg1_alpha = jnp.where(neq_neg1_mask, alpha, 0.0)

    def cdf_when_alpha_neq_neg1():
        one_more_alpha = 1.0 + neq_neg1_alpha
        low_pow_one_more_alpha = jnp.power(low, one_more_alpha)
        return (jnp.power(x, one_more_alpha) - low_pow_one_more_alpha) / (
            jnp.power(high, one_more_alpha) - low_pow_one_more_alpha
        )

    def cdf_when_alpha_eq_neg1():
        return jnp.log(x / low) / jnp.log(high / low)

    cdf_val = jnp.where(
        neq_neg1_mask,
        cdf_when_alpha_neq_neg1(),
        cdf_when_alpha_eq_neg1(),
    )
    return jnp.clip(cdf_val, 0.0, 1.0)


@doubly_truncated_power_law_cdf.defjvp
def doubly_truncated_power_law_cdf_jvp(primals, tangents):
    # source https://github.com/pyro-ppl/numpyro/blob/94f4b99710d855bea456210cf91e6e55eeac3926/numpyro/distributions/truncated.py#L567-L661
    x, alpha, low, high = primals
    x_t, alpha_t, low_t, high_t = tangents

    log_low = jnp.log(low)
    log_high = jnp.log(high)
    log_x = jnp.log(x)

    delta_eq_neg1 = 10e-4
    neq_neg1_mask = jnp.not_equal(alpha, -1.0)
    neq_neg1_alpha = jnp.where(neq_neg1_mask, alpha, 0.0)

    # Calculate primal
    primal_out = doubly_truncated_power_law_cdf(*primals)

    # Tangents for alpha not equals -1
    def x_neq_neg1(alpha):
        one_more_alpha = 1.0 + alpha
        return (one_more_alpha * jnp.power(x, alpha)) / (
            jnp.power(high, one_more_alpha) - jnp.power(low, one_more_alpha)
        )

    def alpha_neq_neg1(alpha):
        one_more_alpha = 1.0 + alpha
        low_pow_one_more_alpha = jnp.power(low, one_more_alpha)
        high_pow_one_more_alpha = jnp.power(high, one_more_alpha)
        x_pow_one_more_alpha = jnp.power(x, one_more_alpha)
        term1 = (x_pow_one_more_alpha * log_x - low_pow_one_more_alpha * log_low) / (
            high_pow_one_more_alpha - low_pow_one_more_alpha
        )
        term2 = (
            (x_pow_one_more_alpha - low_pow_one_more_alpha)
            * (high_pow_one_more_alpha * log_high - low_pow_one_more_alpha * log_low)
        ) / jnp.square(high_pow_one_more_alpha - low_pow_one_more_alpha)
        return term1 - term2

    def low_neq_neg1(alpha):
        one_more_alpha = 1.0 + alpha
        low_pow_one_more_alpha = jnp.power(low, one_more_alpha)
        high_pow_one_more_alpha = jnp.power(high, one_more_alpha)
        x_pow_one_more_alpha = jnp.power(x, one_more_alpha)
        change = high_pow_one_more_alpha - low_pow_one_more_alpha
        term2 = one_more_alpha * jnp.power(low, alpha) / change
        term1 = term2 * (x_pow_one_more_alpha - low_pow_one_more_alpha) / change
        return term1 - term2

    def high_neq_neg1(alpha):
        one_more_alpha = 1.0 + alpha
        low_pow_one_more_alpha = jnp.power(low, one_more_alpha)
        high_pow_one_more_alpha = jnp.power(high, one_more_alpha)
        x_pow_one_more_alpha = jnp.power(x, one_more_alpha)
        return -(
            one_more_alpha
            * jnp.power(high, alpha)
            * (x_pow_one_more_alpha - low_pow_one_more_alpha)
        ) / jnp.square(high_pow_one_more_alpha - low_pow_one_more_alpha)

    # Tangents for alpha equals -1
    def x_eq_neg1():
        return jnp.reciprocal(x * (log_high - log_low))

    def low_eq_neg1():
        return (log_x - log_low) / (
            jnp.square(log_high - log_low) * low
        ) - jnp.reciprocal((log_high - log_low) * low)

    def high_eq_neg1():
        return (log_x - log_low) / (jnp.square(log_high - log_low) * high)

    # Including approximation for alpha = -1
    tangent_out = (
        jnp.where(neq_neg1_mask, x_neq_neg1(neq_neg1_alpha), x_eq_neg1()) * x_t
        + jnp.where(
            neq_neg1_mask,
            alpha_neq_neg1(neq_neg1_alpha),
            (
                alpha_neq_neg1(alpha - delta_eq_neg1)
                + alpha_neq_neg1(alpha + delta_eq_neg1)
            )
            * 0.5,
        )
        * alpha_t
        + jnp.where(neq_neg1_mask, low_neq_neg1(neq_neg1_alpha), low_eq_neg1()) * low_t
        + jnp.where(neq_neg1_mask, high_neq_neg1(neq_neg1_alpha), high_eq_neg1())
        * high_t
    )

    return primal_out, tangent_out


@jax.custom_jvp
def doubly_truncated_power_law_icdf(q, alpha, low, high):
    # source https://github.com/pyro-ppl/numpyro/blob/94f4b99710d855bea456210cf91e6e55eeac3926/numpyro/distributions/truncated.py#L680-L703
    neq_neg1_mask = jnp.not_equal(alpha, -1.0)
    neq_neg1_alpha = jnp.where(neq_neg1_mask, alpha, 0.0)

    def icdf_alpha_neq_neg1():
        one_more_alpha = 1.0 + neq_neg1_alpha
        low_pow_one_more_alpha = jnp.power(low, one_more_alpha)
        high_pow_one_more_alpha = jnp.power(high, one_more_alpha)
        return jnp.power(
            low_pow_one_more_alpha
            + q * (high_pow_one_more_alpha - low_pow_one_more_alpha),
            jnp.reciprocal(one_more_alpha),
        )

    def icdf_alpha_eq_neg1():
        return jnp.power(high / low, q) * low

    icdf_val = jnp.where(
        neq_neg1_mask,
        icdf_alpha_neq_neg1(),
        icdf_alpha_eq_neg1(),
    )
    return icdf_val


@doubly_truncated_power_law_icdf.defjvp
def doubly_truncated_power_law_icdf_jvp(primals, tangents):
    # source https://github.com/pyro-ppl/numpyro/blob/94f4b99710d855bea456210cf91e6e55eeac3926/numpyro/distributions/truncated.py#L705-L815
    x, alpha, low, high = primals
    x_t, alpha_t, low_t, high_t = tangents

    log_low = jnp.log(low)
    log_high = jnp.log(high)
    high_over_low = jnp.divide(high, low)

    delta_eq_neg1 = 10e-4
    neq_neg1_mask = jnp.not_equal(alpha, -1.0)
    neq_neg1_alpha = jnp.where(neq_neg1_mask, alpha, 0.0)

    primal_out = doubly_truncated_power_law_icdf(*primals)

    # Tangents for alpha not equal -1
    def x_neq_neg1(alpha):
        one_more_alpha = 1.0 + alpha
        low_pow_one_more_alpha = jnp.power(low, one_more_alpha)
        high_pow_one_more_alpha = jnp.power(high, one_more_alpha)
        change = high_pow_one_more_alpha - low_pow_one_more_alpha
        return (
            change
            * jnp.power(
                low_pow_one_more_alpha + x * change,
                jnp.reciprocal(one_more_alpha) - 1,
            )
        ) / one_more_alpha

    def alpha_neq_neg1(alpha):
        one_more_alpha = 1.0 + alpha
        low_pow_one_more_alpha = jnp.power(low, one_more_alpha)
        high_pow_one_more_alpha = jnp.power(high, one_more_alpha)
        factor0 = low_pow_one_more_alpha + x * (
            high_pow_one_more_alpha - low_pow_one_more_alpha
        )
        term1 = jnp.power(factor0, jnp.reciprocal(one_more_alpha))
        term2 = (
            low_pow_one_more_alpha * log_low
            + x
            * (high_pow_one_more_alpha * log_high - low_pow_one_more_alpha * log_low)
        ) / (one_more_alpha * factor0)
        term3 = jnp.log(factor0) / jnp.square(one_more_alpha)
        return term1 * (term2 - term3)

    def low_neq_neg1(alpha):
        one_more_alpha = 1.0 + alpha
        low_pow_one_more_alpha = jnp.power(low, one_more_alpha)
        high_pow_one_more_alpha = jnp.power(high, one_more_alpha)
        return (
            (1.0 - x)
            * jnp.power(low, alpha)
            * jnp.power(
                low_pow_one_more_alpha
                + x * (high_pow_one_more_alpha - low_pow_one_more_alpha),
                jnp.reciprocal(one_more_alpha) - 1,
            )
        )

    def high_neq_neg1(alpha):
        one_more_alpha = 1.0 + alpha
        low_pow_one_more_alpha = jnp.power(low, one_more_alpha)
        high_pow_one_more_alpha = jnp.power(high, one_more_alpha)
        return (
            x
            * jnp.power(high, alpha)
            * jnp.power(
                low_pow_one_more_alpha
                + x * (high_pow_one_more_alpha - low_pow_one_more_alpha),
                jnp.reciprocal(one_more_alpha) - 1,
            )
        )

    # Tangents for alpha equals -1
    def dx_eq_neg1():
        return low * jnp.power(high_over_low, x) * (log_high - log_low)

    def low_eq_neg1():
        return (
            jnp.power(high_over_low, x)
            - (high * x * jnp.power(high_over_low, x - 1)) / low
        )

    def high_eq_neg1():
        return x * jnp.power(high_over_low, x - 1)

    # Including approximation for alpha = -1 \
    tangent_out = (
        jnp.where(neq_neg1_mask, x_neq_neg1(neq_neg1_alpha), dx_eq_neg1()) * x_t
        + jnp.where(
            neq_neg1_mask,
            alpha_neq_neg1(neq_neg1_alpha),
            (
                alpha_neq_neg1(alpha - delta_eq_neg1)
                + alpha_neq_neg1(alpha + delta_eq_neg1)
            )
            * 0.5,
        )
        * alpha_t
        + jnp.where(neq_neg1_mask, low_neq_neg1(neq_neg1_alpha), low_eq_neg1()) * low_t
        + jnp.where(neq_neg1_mask, high_neq_neg1(neq_neg1_alpha), high_eq_neg1())
        * high_t
    )

    return primal_out, tangent_out



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


class ModelRegistry(object):
    r"""Decorator class to register models and their parameters."""

    def __init__(self) -> None:
        self._registry = {}

    @property
    def registry(self) -> dict:
        r"""Hashmap of registered models and their parameters."""
        return self._registry

    def register(self, parameter, model=None):
        r"""Registers a model with the parameter(s) it yields."""
        if model is None:
            return lambda model: self.register(parameter, model)
        if isinstance(parameter, str):
            parameter = (parameter,)
        elif isinstance(parameter, tuple):
            assert all(isinstance(p, str) for p in parameter)
        else:
            raise ValueError("Parameter must be a string or tuple of strings")

        self._registry[parameter] = model
        return model

    def __call__(self, parameter):
        try:
            model = self._registry[parameter]
        except KeyError as e:
            raise NotImplementedError from e

        return model(parameter)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from ._doubletruncpowerlaw import (
    doubly_truncated_power_law_cdf as doubly_truncated_power_law_cdf,
    doubly_truncated_power_law_icdf as doubly_truncated_power_law_icdf,
    doubly_truncated_power_law_log_norm_constant as doubly_truncated_power_law_log_norm_constant,
    doubly_truncated_power_law_log_prob as doubly_truncated_power_law_log_prob,
)
from ._extendedsupporttransformeddistribution import (
    ExtendedSupportTransformedDistribution as ExtendedSupportTransformedDistribution,
)
from ._joindistribution import JointDistribution as JointDistribution
from ._scaledmixture import ScaledMixture as ScaledMixture



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import os
import warnings
from collections.abc import Callable
from typing import List, Optional, Tuple, Union

import h5py
import numpy as np
import tqdm
from jax import nn as jnn, numpy as jnp, random as jrd
from jaxtyping import Array, PRNGKeyArray
from loguru import logger
from numpyro.util import is_prng_key

from ..models.utils import ScaledMixture
from ..models.wrappers import ModelRegistry
from ..poisson_mean import PoissonMean
from ..utils.tools import error_if
from ._utils import ensure_dat_extension


PROGRESS_BAR_TEXT_WITDH = 25

error_magazine = ModelRegistry()


class PopulationFactory:
    error_dir: str = "posteriors"
    event_filename: str = "event_{}"
    injection_filename: str = "injections"
    realizations_dir: str = "realization_{}"
    root_dir: str = "data"
    verbose: bool = True
    mean_covs_filename: str = "means_covs.hdf5"

    def __init__(
        self,
        model_fn: Union[ScaledMixture, Callable[..., ScaledMixture]],
        model_params: dict[str, Array],
        parameters: List[str],
        log_selection_fn: Optional[Callable[[Array], Array]],
        ERate_obj: PoissonMean,
        num_realizations: int = 5,
        error_size: int = 2_000,
    ) -> None:
        """Class with methods equipped to generate population for each realizations and
        adding errors in it.

        Parameters
        ----------
        model_fn : Union[ScaledMixture, Callable[..., ScaledMixture]]
            Model for the population. If a callable is provided, it should return a
            `ScaledMixture` model.
        model_params : dict[str, Array]
            Parameters for the model.
        parameters : List[str]
            Parameters for the model in order.
        log_selection_fn : Callable[[Array], Array]
            logarithm of volume time sensitivity function.
        ERate_fn : Callable[[ScaledMixture, Optional[int]], Array]
            Expected rate function.
        num_realizations : int, optional
            Number of realizations to generate, by default 5
        error_size : int, optional
            Size of the error to add in the population, by default 2_000

        Raises
        ------
        ValueError
            If model is not provided.
        ValueError
            If provided model is not a `ScaledMixture` model.
        ValueError
            If parameters are not provided.
        """
        self.model: ScaledMixture = model_fn(**model_params)
        self.parameters = parameters
        self.log_selection_fn = log_selection_fn
        self.ERate_fn = ERate_obj
        self.num_realizations = num_realizations
        self.error_size = error_size

        self.event_filename = ensure_dat_extension(self.event_filename)
        self.injection_filename = ensure_dat_extension(self.injection_filename)

    def _generate_population(
        self, size: int, *, key: PRNGKeyArray
    ) -> Tuple[Array, Array, Array, Array]:
        r"""Generate population for a realization."""

        old_size = size
        if self.log_selection_fn is not None:
            size += int(1e4)

        population, [indices] = self.model.sample_with_intermediates(key, (size,))

        raw_population = population
        raw_indices = indices

        m1_index = self.parameters.index("mass_1_source")
        m2_index = self.parameters.index("mass_2_source")

        m1 = population[:, m1_index]
        m2 = population[:, m2_index]

        mask = np.less_equal(m2, m1)

        count = jnp.sum(mask)

        logger.debug(
            "Number of injections with m2 <= m1: {count} out of {size}",
            count=count,
            size=size,
        )

        population = population[mask]
        indices = indices[mask]

        if self.log_selection_fn is not None:
            _, key = jrd.split(key)
            log_selection = self.log_selection_fn(population)
            log_selection = jnp.nan_to_num(
                log_selection,
                nan=-jnp.inf,
                posinf=-jnp.inf,
                neginf=-jnp.inf,
            )
            vt = jnn.softmax(log_selection)
            _, key = jrd.split(key)
            index = jrd.choice(
                key, jnp.arange(population.shape[0]), p=vt, shape=(old_size,)
            )

            population = population[index]
            indices = indices[index]

        return raw_population, raw_indices, population, indices

    def _generate_realizations(self, key: PRNGKeyArray) -> None:
        r"""Generate realizations for the population."""
        poisson_key, rate_key = jrd.split(key)
        exp_rate = self.ERate_fn(self.model)
        logger.debug(f"Expected rate for the population is {exp_rate}")
        size = int(jrd.poisson(poisson_key, exp_rate))
        logger.debug(f"Population size is {size}")
        key = rate_key
        if size <= 0:
            raise ValueError(
                f"Population size is {size}. This can be a result of following:\n"
                "\t1. The rate is zero.\n"
                "\t2. The volume is zero.\n"
                "\t3. The models are not selected for rate calculation.\n"
                "\t4. VT file is not provided or is not valid.\n"
                "\t5. Or some other reason."
            )
        pop_keys = jrd.split(key, self.num_realizations)
        os.makedirs(self.root_dir, exist_ok=True)
        with tqdm.tqdm(
            range(self.num_realizations),
            unit="realization",
            total=self.num_realizations,
        ) as pbar:
            pbar.set_description(
                "Generating population".ljust(PROGRESS_BAR_TEXT_WITDH, " ")
            )
            for i in pbar:
                realizations_path = os.path.join(
                    self.root_dir, self.realizations_dir.format(i)
                )
                os.makedirs(realizations_path, exist_ok=True)

                raw_population, raw_indices, population, indices = (
                    self._generate_population(size, key=pop_keys[i])
                )

                if population.shape == ():
                    continue

                injections_file_path = os.path.join(
                    realizations_path, self.injection_filename
                )
                color_indices_file_path = os.path.join(realizations_path, "color.dat")
                np.savetxt(
                    injections_file_path,
                    population,
                    header=" ".join(self.parameters),
                    comments="",  # To remove the default comment character '#'
                )
                np.savetxt(
                    color_indices_file_path,
                    indices,
                    comments="",  # To remove the default comment character '#'
                    fmt="%d",
                )
                raw_injections_file_path = os.path.join(
                    realizations_path, "raw_" + self.injection_filename
                )
                raw_color_indices_file_path = os.path.join(
                    realizations_path, "raw_color.dat"
                )
                np.savetxt(
                    raw_injections_file_path,
                    raw_population,
                    header=" ".join(self.parameters),
                    comments="",  # To remove the default comment character '#'
                )
                np.savetxt(
                    raw_color_indices_file_path,
                    raw_indices,
                    comments="",  # To remove the default comment character '#'
                    fmt="%d",
                )

    def _add_error(self, realization_number, *, key: PRNGKeyArray) -> None:
        r"""Adds error to the realizations' population."""
        realizations_path = os.path.join(
            self.root_dir, self.realizations_dir.format(realization_number)
        )

        heads: List[List[int]] = []
        error_fns: List[Callable[[Array, int, PRNGKeyArray], Array]] = []

        for head, err_fn in error_magazine.registry.items():
            _head = []
            for h in head:
                i = self.parameters.index(h)
                _head.append(i)
            heads.append(_head)
            error_fns.append(err_fn)

        output_dir = os.path.join(
            realizations_path, self.error_dir, self.event_filename
        )

        os.makedirs(os.path.join(realizations_path, self.error_dir), exist_ok=True)

        injections_file_path = os.path.join(realizations_path, self.injection_filename)
        data_inj = np.loadtxt(injections_file_path, skiprows=1)

        n_injections = data_inj.shape[0]

        keys = jrd.split(key, n_injections * (len(heads) + 1))

        means: List[Optional[np.ndarray]] = [None for _ in range(n_injections)]
        covs: List[Optional[np.ndarray]] = [None for _ in range(n_injections)]

        for index in range(n_injections):
            noisy_data = np.empty((self.error_size, len(self.parameters)))
            data = data_inj[index]
            i = 0
            for head, err_fn in zip(heads, error_fns):
                key_idx = index * len(heads) + i
                noisy_data_i: Array = err_fn(data[head], self.error_size, keys[key_idx])
                if noisy_data_i.ndim == 1:
                    noisy_data_i = noisy_data_i.reshape(self.error_size, -1)
                noisy_data[:, head] = noisy_data_i
                i += 1
            nan_mask = np.isnan(noisy_data).any(axis=1)
            noisy_data = noisy_data[~nan_mask]  # type: ignore

            if noisy_data.shape[0] < 1:
                warnings.warn(
                    f"Skipping file {index} due to all NaN values or insufficient data.",
                    category=UserWarning,
                )
                continue

            mean = np.mean(noisy_data, axis=0)
            error_if(
                np.isnan(mean).any(),
                msg=f"Mean for realization {realization_number} contains NaN values.",
            )

            cov = np.cov(noisy_data.T)
            error_if(
                np.isnan(cov).any(),
                msg=f"Covariance for realization {realization_number} contains NaN values.",
            )

            means[index] = mean
            covs[index] = cov

            np.savetxt(
                output_dir.format(index),
                noisy_data,
                header=" ".join(self.parameters),
                comments="",  # To remove the default comment character '#'
            )

        with h5py.File(
            os.path.join(realizations_path, self.mean_covs_filename), "w"
        ) as f:
            for i, (mean, cov) in enumerate(zip(means, covs)):
                if mean is not None and cov is not None:
                    event_name = "event_{}".format(i)
                    event_group = f.create_group(event_name)
                    event_group.create_dataset("mean", data=mean)
                    event_group.create_dataset("cov", data=cov)

    def produce(self, key: PRNGKeyArray) -> None:
        """Generate realizations and add errors to the populations.

        Parameters
        ----------
        key : PRNGKeyArray
            Pseudo-random number generator key for reproducibility, by default None
        """
        assert is_prng_key(key)
        self._generate_realizations(key)
        keys = jrd.split(key, self.num_realizations)
        with tqdm.tqdm(
            range(self.num_realizations),
            unit="realization",
            total=self.num_realizations,
        ) as pbar:
            pbar.set_description("Adding errors".ljust(PROGRESS_BAR_TEXT_WITDH, " "))
            for i in pbar:
                self._add_error(i, key=keys[i])



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from typing import List

from gwkokab.parameters import Parameters
from kokab.ecc_matters.common import EccentricityMattersModel
from kokab.utils.f_monk import Monk, monk_arg_parser


class EccentricityMattersMonk(Monk):
    @property
    def parameters(self) -> List[str]:
        return [
            Parameters.PRIMARY_MASS_SOURCE.value,
            Parameters.SECONDARY_MASS_SOURCE.value,
            Parameters.ECCENTRICITY.value,
        ]

    @property
    def model_parameters(self) -> List[str]:
        return ["log_rate", "alpha_m", "mmin", "mmax", "loc", "scale", "low", "high"]


def main() -> None:
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = monk_arg_parser(parser)

    args = parser.parse_args()

    EccentricityMattersMonk(
        model=EccentricityMattersModel,
        data_filename=args.data_filename,
        seed=args.seed,
        prior_filename=args.prior_json,
        selection_fn_filename=args.vt_json,
        poisson_mean_filename=args.pmean_json,
        sampler_settings_filename=args.sampler_config,
        debug_nans=args.debug_nans,
        profile_memory=args.profile_memory,
        check_leaks=args.check_leaks,
        analysis_name="ecc_matters",
        n_samples=args.n_samples,
        max_iter_mean=args.max_iter_mean,
        max_iter_cov=args.max_iter_cov,
        n_vi_steps=args.n_vi_steps,
        learning_rate=args.learning_rate,
        batch_size=args.batch_size,
    ).run()



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import argparse
import os

import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt


def make_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Joint plot",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        epilog="This script plots a joint plot of two columns of a data file.",
    )
    parser.add_argument(
        "-d",
        "--data",
        help="data file path. Only .dat files are supported.",
        required=True,
        type=argparse.FileType("r"),
    )
    parser.add_argument(
        "-o",
        "--output",
        help="output file path",
        required=True,
        type=argparse.FileType("w"),
    )
    parser.add_argument(
        "-x",
        "--x-column-name",
        help="column name for x-axis",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-y",
        "--y-column-name",
        help="column name for y-axis",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-xl",
        "--xlabel",
        help="label for x-axis",
        type=str,
    )
    parser.add_argument(
        "-yl",
        "--ylabel",
        help="label for y-axis",
        type=str,
    )
    parser.add_argument(
        "-t",
        "--title",
        help="title of the plot",
        type=str,
    )
    parser.add_argument(
        "-size",
        "--size",
        help="size of the corner plot in inches",
        nargs=2,
        default=(6, 6),
        type=float,
    )
    parser.add_argument(
        "-cmap",
        "--cmap",
        help="color map",
        default="rocket",
        type=str,
    )
    parser.add_argument(
        "-c",
        "--color",
        help="color of the corner plot",
        default="#180F29",
        type=str,
    )
    parser.add_argument(
        "--use-latex",
        help="use LaTeX for rendering text",
        action="store_true",
    )
    parser.add_argument(
        "--font-family",
        help="font family to use",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--dpi",
        help="dots per inch to save file",
        type=int,
        default=100,
    )

    return parser


def main() -> None:
    """Main function of the script."""
    parser = make_parser()
    args = parser.parse_args()

    plt.rcParams.update({"text.usetex": args.use_latex})
    if args.font_family is not None:
        plt.rcParams.update({"font.family": args.font_family})

    data = pd.read_csv(args.data.name, delimiter=" ")

    g = sns.jointplot(
        x=data[args.x_column_name].to_numpy(),
        y=data[args.y_column_name].to_numpy(),
        marginal_ticks=True,
        ratio=2,
    )
    g.plot_marginals(sns.histplot, color=args.color)
    g.plot_joint(sns.kdeplot, fill=True, thresh=0, cmap=args.cmap)
    g.set_axis_labels(args.xlabel, args.ylabel)
    g.fig.suptitle(args.title)
    # Determine output file type and save accordingly
    output_ext = os.path.splitext(args.output.name)[1].lower()
    plt_savefig_kwargs = dict()
    if output_ext == ".png":
        plt_savefig_kwargs["dpi"] = args.dpi
    g.savefig(args.output.name, bbox_inches="tight", **plt_savefig_kwargs)
    g.close("all")



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser

import pandas as pd

from gwkokab.parameters import Parameters
from gwkokab.utils.tools import error_if
from kokab.ecc_matters.common import EccentricityMattersModel
from kokab.utils import ppd, ppd_parser
from kokab.utils.common import ppd_ranges, read_json


def make_parser() -> ArgumentParser:
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = ppd_parser.get_parser(parser)

    return parser


def main() -> None:
    parser = make_parser()
    args = parser.parse_args()

    error_if(
        not str(args.filename).endswith(".hdf5"),
        msg="Output file must be an HDF5 file.",
    )

    constants = read_json(args.constants)
    nf_samples_mapping = read_json(args.nf_samples_mapping)

    parameters = [
        Parameters.PRIMARY_MASS_SOURCE.value,
        Parameters.SECONDARY_MASS_SOURCE.value,
        Parameters.ECCENTRICITY.value,
    ]
    ranges = ppd_ranges(parameters, args.range)

    nf_samples = pd.read_csv(
        args.sample_filename, delimiter=" ", comment="#", header=None
    ).to_numpy()

    ppd.compute_and_save_ppd(
        EccentricityMattersModel,
        nf_samples,
        ranges,
        "rate_scaled_" + args.filename,
        parameters,
        constants,
        nf_samples_mapping,
        args.batch_size,
    )

    nf_samples, constants = ppd.wipe_log_rate(nf_samples, nf_samples_mapping, constants)

    ppd.compute_and_save_ppd(
        EccentricityMattersModel,
        nf_samples,
        ranges,
        args.filename,
        parameters,
        constants,
        nf_samples_mapping,
        args.batch_size,
    )



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0

#
"""All necessary constants for the package."""

from typing import Final


SPEED_OF_LIGHT: Final[float] = 299792458.0
r"""Speed of light in vacuum in :math:`\text{m}/\text{s}`"""
MSUN_SI: Final[float] = 1.9884099021470415e30
"""Solar mass in kg."""
SECONDS_PER_YEAR: Final[float] = 365.25 * 24 * 60 * 60
"""Seconds in 1 year."""



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from argparse import ArgumentParser

from numpyro.distributions.distribution import enable_validation


def get_parser(parser: ArgumentParser) -> ArgumentParser:
    """Populate the command line argument parser with the arguments for the Genie
    script.

    Parameters
    ----------
    parser : ArgumentParser
        Parser to add the arguments to

    Returns
    -------
    ArgumentParser
        the command line argument parser
    """

    # Global enable validation for all distributions
    enable_validation()

    genie_group = parser.add_argument_group("Genie Options")

    genie_group.add_argument(
        "--error-size",
        help="Size of the error.",
        default=2000,
        type=int,
    )
    genie_group.add_argument(
        "--num-realizations",
        help="Number of realizations.",
        default=5,
        type=int,
    )
    genie_group.add_argument(
        "--seed",
        help="Random seed for reproducibility.",
        default=37,
        type=int,
    )

    vt_group = parser.add_argument_group("VT Options")

    vt_group.add_argument(
        "--vt-json",
        help="Path to the JSON file containing the VT options.",
        type=str,
        default="vt.json",
    )

    pmean_group = parser.add_argument_group("Poisson Mean Options")

    pmean_group.add_argument(
        "--pmean-json",
        help="Path to the JSON file containing the Poisson mean options.",
        type=str,
        default="pmean.json",
    )

    return parser



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from collections.abc import Callable, Sequence
from typing import Literal, Optional, overload, Tuple

import equinox as eqx
import h5py
import jax
from jax import numpy as jnp
from jax.numpy import trapezoid
from jax.scipy.interpolate import RegularGridInterpolator
from jaxtyping import Array

from gwkokab.parameters import Parameters
from gwkokab.utils.transformations import (
    chirp_mass,
    log_chirp_mass,
    symmetric_mass_ratio,
)

from ._abc import VolumeTimeSensitivityInterface


def _check_and_get(name: str, f: h5py.File) -> Array:
    """Check if the name is in the file and return the value.

    Parameters
    ----------
    name : str
        name of the dataset
    f : h5py.File
        h5py file object

    Returns
    -------
    Array
        the value of the dataset

    Raises
    ------
    ValueError
        if the name is not in the file
    """
    if name in f:
        return f[name][()]
    else:
        raise ValueError(f"{name} not found in file.")


class PopModelsVolumeTimeSensitivity(VolumeTimeSensitivityInterface):
    logVT_interpolator: RegularGridInterpolator
    """Interpolator for the log volume time sensitivity function."""
    m_min: float = eqx.field(converter=float, static=True)
    """Minimum mass."""

    def __init__(
        self,
        parameters: Sequence[str],
        filename: str,
        zero_spin: bool,
        scale_factor: int = 1,
        m_min: float = 0.5,
        batch_size: Optional[int] = None,
    ) -> None:
        r"""Convenience class for loading a volume time sensitivity function generated
        by `PopModels <https://gitlab.com/dwysocki/bayesian-parametric-population-models>`_.

        parameters : Sequence[str]
            The names of the parameters that the model expects.
        filename : str
            The filename of the volume time sensitivity function.
        zero_spin : bool
            Load with zero spin or not. :code:`True` for zero spin, :code:`False` otherwise.
        scale_factor : int
            Scale factor for the volume time sensitivity function, by default 1
        m_min : float
            Minimum mass, by default 0.5
        batch_size : Optional[int], optional
            The batch size :func:`jax.lax.map` should use, by default None.
        """
        if batch_size is not None:
            if not isinstance(batch_size, int):
                raise TypeError(
                    f"batch_size must be an integer, got {type(batch_size)}"
                )
            if batch_size < 1:
                raise ValueError(
                    f"batch_size must be a positive integer, got {batch_size}"
                )

        self.batch_size = batch_size
        self.m_min = m_min
        if (
            Parameters.PRIMARY_MASS_SOURCE.value not in parameters
            or Parameters.SECONDARY_MASS_SOURCE.value not in parameters
        ):
            raise ValueError(
                f"{self.__class__.__name__} requires "
                f"{Parameters.PRIMARY_MASS_SOURCE.value} and "
                f"{Parameters.SECONDARY_MASS_SOURCE.value}"
            )
        spin_required = (
            Parameters.PRIMARY_SPIN_MAGNITUDE.value in parameters
            and Parameters.SECONDARY_SPIN_MAGNITUDE.value in parameters
        )

        if not spin_required and not zero_spin:
            VT, logM, qtilde, s1z, s2z = self._load_file(filename, zero_spin)
            VT = trapezoid(VT, s2z, axis=-1)
            VT = trapezoid(VT, s1z, axis=-1)
            del s1z, s2z
            points = (logM, qtilde)
            self.shuffle_indices = [
                parameters.index(Parameters.PRIMARY_MASS_SOURCE.value),
                parameters.index(Parameters.SECONDARY_MASS_SOURCE.value),
            ]
        elif spin_required and not zero_spin:
            VT, logM, qtilde, s1z, s2z = self._load_file(filename, zero_spin)
            points = (logM, qtilde, s1z, s2z)
            self.shuffle_indices = [
                parameters.index(Parameters.PRIMARY_MASS_SOURCE.value),
                parameters.index(Parameters.SECONDARY_MASS_SOURCE.value),
                parameters.index(Parameters.PRIMARY_SPIN_MAGNITUDE.value),
                parameters.index(Parameters.SECONDARY_SPIN_MAGNITUDE.value),
            ]
        elif not spin_required and zero_spin:
            VT, logM, qtilde = self._load_file(filename, zero_spin)
            points = (logM, qtilde)
            self.shuffle_indices = [
                parameters.index(Parameters.PRIMARY_MASS_SOURCE.value),
                parameters.index(Parameters.SECONDARY_MASS_SOURCE.value),
            ]
        else:
            raise ValueError(
                "Spin is required by the model, but zero_spin is set to True."
            )

        logVT = jnp.log(VT * scale_factor)

        self.logVT_interpolator = RegularGridInterpolator(
            points, logVT, bounds_error=False, fill_value=-jnp.inf
        )

    @overload
    @staticmethod
    def _load_file(
        self, filename: str, zero_spin: Literal[False]
    ) -> Tuple[Array, Array, Array, Array, Array]: ...

    @overload
    @staticmethod
    def _load_file(
        self, filename: str, zero_spin: Literal[True]
    ) -> Tuple[Array, Array, Array]: ...

    @staticmethod
    def _load_file(filename: str, zero_spin: bool):
        with h5py.File(filename, "r") as f:
            VT = _check_and_get("VT", f)
            logM = _check_and_get("logM", f)
            qtilde = _check_and_get("qtilde", f)
            s1z = _check_and_get("s1z", f)
            s2z = _check_and_get("s2z", f)

            if zero_spin:
                # only retain axis where s1z == s2z == 0
                s1z_idx = s1z == 0
                s2z_idx = s2z == 0
                VT = jnp.squeeze(VT[..., s1z_idx, s2z_idx], axis=-1)

        if zero_spin:
            return VT, logM, qtilde
        else:
            return VT, logM, qtilde, s1z, s2z

    @eqx.filter_jit
    def mass_grid_coords_inverse(
        self, logM: Array, q_tilde: Array
    ) -> Tuple[Array, Array]:
        r"""Converts the :math:`\log{(M)}`, :math:`\tilde{q}` coordinates to m1, m2
        masses.

        .. math::
            m_1 = \frac{M(M-m_{\text{min}})}{M(1+\tilde{q}) - 2m_{\text{min}}\tilde{q}}

        .. math::
            m_2 = \frac{M(M\tilde{q} + m_{\text{min}}(1-2\tilde{q}))}{M(1+\tilde{q}) - 2m_{\text{min}}\tilde{q}}

        where :math:`M = \exp{(\log{(M)})}`. `source <https://gitlab.com/dwysocki/bayesian-parametric-population-models/-/blob/master/src/pop_models/astro_models/gw_ifo_vt.py?ref_type=heads#L1025-1037>`_.

        Parameters
        ----------
        logM : Array
            Logarithm of the total mass
        q_tilde : Array
            Reduced mass ratio

        Returns
        -------
        Tuple[Array, Array]
            Primary and secondary masses
        """
        M = jnp.exp(logM)
        denominator = M * (1 + q_tilde) - 2 * self.m_min * q_tilde
        factor = (M - self.m_min) / denominator
        m1 = M * factor
        m2 = M - m1
        return m1, m2

    @eqx.filter_jit
    def mass_grid_coords(
        self, m1: Array, m2: Array, eps: float = 1e-8
    ) -> Tuple[Array, Array]:
        r"""Converts the :math:`m_1`, :math:`m_2` masses to :math:`\log{(M)}`,
        :math:`\tilde{q}` coordinates.

        .. note::

            When computing :math:`\tilde{q}`, there is a coordinate singularity at
            :math:`m_1 = m_2 = m_{\text{min}}`. While normally, equal mass corresponds to
            :math:`\tilde{q} = 1`, here the most extreme mass ratio is also equal mass, so
            we should have :math:`\tilde{q}` = 0. Here we break the singularity by just
            forcing :math:`\tilde{q} = 1` in these cases. We also allow for some numerical
            wiggle room beyond that point, by also including any points within some small
            epsilon of the singularity. VT does not change on a fast enough scale for this
            to make any measurable difference.

        `source <https://gitlab.com/dwysocki/bayesian-parametric-population-models/-/blob/master/src/pop_models/astro_models/gw_ifo_vt.py?ref_type=heads#L998-1022>`_

        Parameters
        ----------
        m1 : Array
            Primary mass
        m2 : Array
            Secondary mass
        eps : float
            tolerance for numerical stability, defaults to 1e-8

        Returns
        -------
        Tuple[Array, Array]
            coordinates in the form of :math:`\log{(M)}`, :math:`\tilde{q}`
        """
        M = m1 + m2

        logM = jnp.log(M)

        i_good = M > 2 * self.m_min + eps

        # ~i_good is filled with random values
        m1 = jnp.where(i_good, m1, 1.0)
        m2 = jnp.where(i_good, m2, 2.0)
        M = jnp.where(i_good, M, 1.0)

        qtilde = jnp.where(
            i_good, (M * (m2 - self.m_min)) / (m1 * (M - 2 * self.m_min)), 1.0
        )

        return logM, qtilde

    def get_logVT(self) -> Callable[[Array], Array]:
        @eqx.filter_jit
        def _logVT(x: Array) -> Array:
            xs = x[..., self.shuffle_indices]
            m1, m2 = xs[..., 0], xs[..., 1]
            logM, qtilde = self.mass_grid_coords(m1, m2)
            query = (logM, qtilde, *xs[..., 2:])
            return self.logVT_interpolator(query)

        return _logVT

    def get_mapped_logVT(self) -> Callable[[Array], Array]:
        _batch_size = self.batch_size
        return lambda x: jax.lax.map(self.get_logVT(), x, batch_size=_batch_size)


# source: https://gitlab.com/dwysocki/bayesian-parametric-population-models/-/blob/master/src/pop_models/astro_models/gw_ifo_vt.py?ref_type=heads#L554-709

_correction_basis_scalar_zero_spin = [
    lambda m1, m2: 1.0,
]
_correction_basis_scalar_aligned_spin = [
    lambda m1, m2, a1z, a2z: 1.0,
]

_correction_basis_linear_zero_spin = _correction_basis_scalar_zero_spin + [
    lambda m1, m2: m1,
    lambda m1, m2: m2,
]
_correction_basis_linear_aligned_spin = _correction_basis_scalar_aligned_spin + [
    lambda m1, m2, a1z, a2z: m1,
    lambda m1, m2, a1z, a2z: m2,
]

_correction_basis_quadratic_zero_spin = _correction_basis_linear_zero_spin + [
    lambda m1, m2: m1 * m2,
    lambda m1, m2: m1**2,
    lambda m1, m2: m2**2,
]
_correction_basis_quadratic_aligned_spin = _correction_basis_linear_aligned_spin + [
    lambda m1, m2, a1z, a2z: m1 * m2,
    lambda m1, m2, a1z, a2z: m1**2,
    lambda m1, m2, a1z, a2z: m2**2,
]
_correction_basis_quintic_zero_spin = _correction_basis_quadratic_zero_spin + [
    lambda m1, m2: m1**2 * m2,
    lambda m1, m2: m1 * m2**2,
    lambda m1, m2: m1**3,
    lambda m1, m2: m2**3,
]
_correction_basis_quintic_aligned_spin = _correction_basis_quadratic_aligned_spin + [
    lambda m1, m2, a1z, a2z: m1**2 * m2,
    lambda m1, m2, a1z, a2z: m1 * m2**2,
    lambda m1, m2, a1z, a2z: m1**3,
    lambda m1, m2, a1z, a2z: m2**3,
]


_correction_basis_linear_Mc_eta_zero_spin = _correction_basis_scalar_zero_spin + [
    lambda m1, m2: chirp_mass(m1, m2),
    lambda m1, m2: symmetric_mass_ratio(m1, m2),
]
_correction_basis_linear_Mc_eta_aligned_spin = _correction_basis_scalar_aligned_spin + [
    lambda m1, m2, a1z, a2z: chirp_mass(m1, m2),
    lambda m1, m2, a1z, a2z: symmetric_mass_ratio(m1, m2),
]

_correction_basis_quadratic_Mc_eta_zero_spin = _correction_basis_linear_zero_spin + [
    lambda m1, m2: chirp_mass(m1, m2) * symmetric_mass_ratio(m1, m2),
    lambda m1, m2: chirp_mass(m1, m2) ** 2,
    lambda m1, m2: symmetric_mass_ratio(m1, m2) ** 2,
]
_correction_basis_quadratic_Mc_eta_aligned_spin = (
    _correction_basis_linear_aligned_spin
    + [
        lambda m1, m2, a1z, a2z: chirp_mass(m1, m2) * symmetric_mass_ratio(m1, m2),
        lambda m1, m2, a1z, a2z: chirp_mass(m1, m2) ** 2,
        lambda m1, m2, a1z, a2z: symmetric_mass_ratio(m1, m2) ** 2,
    ]
)

_correction_basis_linear_logMc_eta_zero_spin = _correction_basis_scalar_zero_spin + [
    lambda m1, m2: log_chirp_mass(m1, m2),
    lambda m1, m2: symmetric_mass_ratio(m1, m2),
]
_correction_basis_linear_logMc_eta_aligned_spin = (
    _correction_basis_scalar_aligned_spin
    + [
        lambda m1, m2, a1z, a2z: log_chirp_mass(m1, m2),
        lambda m1, m2, a1z, a2z: symmetric_mass_ratio(m1, m2),
    ]
)

_correction_basis_quadratic_logMc_eta_zero_spin = _correction_basis_linear_zero_spin + [
    lambda m1, m2: log_chirp_mass(m1, m2) * symmetric_mass_ratio(m1, m2),
    lambda m1, m2: log_chirp_mass(m1, m2) ** 2,
    lambda m1, m2: symmetric_mass_ratio(m1, m2) ** 2,
]
_correction_basis_quadratic_logMc_eta_aligned_spin = (
    _correction_basis_linear_aligned_spin
    + [
        lambda m1, m2, a1z, a2z: log_chirp_mass(m1, m2) * symmetric_mass_ratio(m1, m2),
        lambda m1, m2, a1z, a2z: log_chirp_mass(m1, m2) ** 2,
        lambda m1, m2, a1z, a2z: symmetric_mass_ratio(m1, m2) ** 2,
    ]
)


_correction_bases_zero_spin = {
    "scalar": _correction_basis_scalar_zero_spin,
    "linear": _correction_basis_linear_zero_spin,
    "quadratic": _correction_basis_quadratic_zero_spin,
    "quintic": _correction_basis_quintic_zero_spin,
    "linear_Mc_eta": _correction_basis_linear_Mc_eta_zero_spin,
    "quadratic_Mc_eta": _correction_basis_quadratic_Mc_eta_zero_spin,
    "linear_logMc_eta": _correction_basis_linear_logMc_eta_zero_spin,
    "quadratic_logMc_eta": _correction_basis_quadratic_logMc_eta_zero_spin,
}
_correction_bases_aligned_spin = {
    "scalar": _correction_basis_scalar_aligned_spin,
    "linear": _correction_basis_linear_aligned_spin,
    "quadratic": _correction_basis_quadratic_aligned_spin,
    "quintic": _correction_basis_quintic_aligned_spin,
    "linear_Mc_eta": _correction_basis_linear_Mc_eta_aligned_spin,
    "quadratic_Mc_eta": _correction_basis_quadratic_Mc_eta_aligned_spin,
    "linear_logMc_eta": _correction_basis_linear_logMc_eta_aligned_spin,
    "quadratic_logMc_eta": _correction_basis_quadratic_logMc_eta_aligned_spin,
}


class PopModelsCalibratedVolumeTimeSensitivity(PopModelsVolumeTimeSensitivity):
    coeffs: Array = eqx.field(converter=jnp.asarray)
    """Coefficients for the basis functions."""
    basis: Sequence[Callable[..., float | Array]]
    """Basis functions to use for the correction."""

    def __init__(
        self,
        parameters: Sequence[str],
        filename: str,
        zero_spin: bool,
        coeffs: Sequence[int | float],
        basis: str,
        scale_factor: int = 1,
        m_min: float = 0.5,
        batch_size: Optional[int] = None,
    ) -> None:
        r"""Convenience class for loading a volume time sensitivity function generated
        by `PopModels <https://gitlab.com/dwysocki/bayesian-parametric-population-models>`_
        with calibrated corrections.

        Parameters
        ----------
        parameters : Sequence[str]
            The names of the parameters that the model expects.
        filename : str
            The filename of the volume time sensitivity function.
        zero_spin : bool
            Load with zero spin or not. :code:`True` for zero spin, :code:`False` otherwise.
        coeffs : Sequence[int | float]
            Coefficients for the basis functions
        basis : str
            Basis functions to use for the correction
        scale_factor : int
            Scale factor for the volume time sensitivity function, by default 1
        m_min : float
            Minimum mass, by default 0.5
        """
        self.coeffs = jnp.asarray(coeffs)
        if zero_spin:
            self.basis = _correction_bases_zero_spin[basis]
        else:
            self.basis = _correction_bases_aligned_spin[basis]
        super(PopModelsCalibratedVolumeTimeSensitivity, self).__init__(
            parameters,
            filename,
            zero_spin,
            scale_factor,
            m_min,
            batch_size=batch_size,
        )

    def get_logVT(self) -> Callable[[Array], Array]:
        @eqx.filter_jit
        def _logVT(x: Array) -> Array:
            xs = x[..., self.shuffle_indices]
            m1, m2 = xs[..., 0], xs[..., 1]
            x_converted = [m1, m2, *xs[..., 2:]]

            # Loop over the basis functions and multiply by the coefficients
            # https://gitlab.com/dwysocki/bayesian-parametric-population-models/-/blob/master/src/pop_models/astro_models/gw_ifo_vt.py?ref_type=heads#L487-492
            basis_values = jnp.array([base(*x_converted) for base in self.basis])
            correction = jnp.dot(self.coeffs, basis_values)

            logM, qtilde = self.mass_grid_coords(m1, m2)
            query = (logM, qtilde, *xs[..., 2:])
            safe_correction = jnp.where(correction < 0, 1.0, correction)
            return jnp.where(
                correction < 0,
                -jnp.inf,
                self.logVT_interpolator(query) + jnp.log(safe_correction),
            )

        return _logVT



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from typing import List

from numpyro.distributions.distribution import enable_validation

from gwkokab.inference import numpyro_poisson_likelihood, poisson_likelihood
from gwkokab.parameters import Parameters
from kokab.ecc_matters.common import EccentricityMattersModel
from kokab.utils.flowMC_based import flowMC_arg_parser, FlowMCBased
from kokab.utils.numpyro_based import numpyro_arg_parser, NumpyroBased
from kokab.utils.sage import Sage, sage_arg_parser as sage_parser


class EccentricityMattersCore(Sage):
    @property
    def parameters(self) -> List[str]:
        return [
            Parameters.PRIMARY_MASS_SOURCE.value,
            Parameters.SECONDARY_MASS_SOURCE.value,
            Parameters.ECCENTRICITY.value,
        ]

    @property
    def model_parameters(self) -> List[str]:
        return ["log_rate", "alpha_m", "mmin", "mmax", "loc", "scale", "low", "high"]


class EccentricityMattersFSage(EccentricityMattersCore, FlowMCBased):
    pass


class EccentricityMattersNSage(EccentricityMattersCore, NumpyroBased):
    pass


def f_main() -> None:
    enable_validation()

    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = sage_parser(parser)
    parser = flowMC_arg_parser(parser)

    args = parser.parse_args()

    EccentricityMattersFSage(
        likelihood_fn=poisson_likelihood,
        model=EccentricityMattersModel,
        posterior_regex=args.posterior_regex,
        posterior_columns=args.posterior_columns,
        seed=args.seed,
        prior_filename=args.prior_json,
        selection_fn_filename=args.vt_json,
        poisson_mean_filename=args.pmean_json,
        sampler_settings_filename=args.sampler_config,
        analysis_name="f_sage_ecc_matters",
        n_buckets=args.n_buckets,
        threshold=args.threshold,
        debug_nans=args.debug_nans,
        profile_memory=args.profile_memory,
        check_leaks=args.check_leaks,
        where_fns=None,
    ).run()


def n_main() -> None:
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = sage_parser(parser)
    parser = numpyro_arg_parser(parser)

    args = parser.parse_args()

    EccentricityMattersNSage(
        likelihood_fn=numpyro_poisson_likelihood,
        model=EccentricityMattersModel,
        posterior_regex=args.posterior_regex,
        posterior_columns=args.posterior_columns,
        seed=args.seed,
        prior_filename=args.prior_json,
        selection_fn_filename=args.vt_json,
        poisson_mean_filename=args.pmean_json,
        sampler_settings_filename=args.sampler_config,
        analysis_name="n_sage_ecc_matters",
        n_buckets=args.n_buckets,
        threshold=args.threshold,
        debug_nans=args.debug_nans,
        profile_memory=args.profile_memory,
        check_leaks=args.check_leaks,
        where_fns=None,
    ).run()



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import sys
from typing import List

from ._abc import VolumeTimeSensitivityInterface
from ._neuralpdet import NeuralNetProbabilityOfDetection
from ._neuralvt import NeuralNetVolumeTimeSensitivity
from ._pdet import pdet_O3
from ._popmodelvt import (
    PopModelsCalibratedVolumeTimeSensitivity,
    PopModelsVolumeTimeSensitivity,
)
from ._semianalyticalinjvt import (
    SemiAnalyticalRealInjectionVolumeTimeSensitivity as SemiAnalyticalRealInjectionVolumeTimeSensitivity,
)
from ._syninjvt import (
    SyntheticInjectionVolumeTimeSensitivity as SyntheticInjectionVolumeTimeSensitivity,
)


def __getattr__(name):
    match name:
        case "NeuralNetVolumeTimeSensitivity":
            return NeuralNetVolumeTimeSensitivity
        case "PopModelsVolumeTimeSensitivity":
            return PopModelsVolumeTimeSensitivity
        case "PopModelsCalibratedVolumeTimeSensitivity":
            return PopModelsCalibratedVolumeTimeSensitivity
        case "SyntheticInjectionVolumeTimeSensitivity":
            return SyntheticInjectionVolumeTimeSensitivity
        case "NeuralNetProbabilityOfDetection":
            return NeuralNetProbabilityOfDetection
        case "SemiAnalyticalRealInjectionVolumeTimeSensitivity":
            return SemiAnalyticalRealInjectionVolumeTimeSensitivity
        case "pdet_O3":
            return pdet_O3
        case _:
            raise AttributeError(f"module {__name__} has no attribute {name}")


# Copyright (c) 2024 Colm Talbot
# SPDX-License-Identifier: MIT


class _Available:
    names: List[str] = [
        "NeuralNetVolumeTimeSensitivity",
        "PopModelsCalibratedVolumeTimeSensitivity",
        "SemiAnalyticalRealInjectionVolumeTimeSensitivity",
        "PopModelsVolumeTimeSensitivity",
        "SyntheticInjectionVolumeTimeSensitivity",
        "NeuralNetProbabilityOfDetection",
        "pdet_O3",
    ]

    def keys(self) -> List[str]:
        return self.names

    def __getitem__(self, key) -> VolumeTimeSensitivityInterface:
        return sys.modules[__name__].__getattr__(key)

    def __repr__(self) -> str:
        return repr(self.keys())


available = _Available()
"""Available volume-time sensitivity models."""



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0

#
"""Kokab namespace provides a comprehensive suite of models and utilities for population
analysis of compact binary coalescences (CBCs).

It includes command-line interfaces (CLIs) for generating mock posterior estimates,
conducting population analyses, and producing visualizations. A detailed list of
available models is accessible at :doc:`/autoapi/gwkokab/models/index`.
"""

import os

from . import (
    ecc_matters as ecc_matters,
    n_pls_m_gs as n_pls_m_gs,
    one_powerlaw_one_peak as one_powerlaw_one_peak,
    utils as utils,
)
from .utils.logger import log_info as _log_info, set_log_level as _set_log_level


_set_log_level(os.environ.get("GWKOKAB_LOG_LEVEL", "TRACE"))
del _set_log_level

_log_info(start=True)
del _log_info



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from typing import Optional

from jax import lax, numpy as jnp, random as jrd
from jax.scipy import special
from jax.scipy.stats import norm, uniform
from jaxtyping import Array, ArrayLike
from numpyro.distributions import constraints, Distribution, Independent
from numpyro.distributions.util import promote_shapes, validate_sample

from ..models.spin import BetaFromMeanVar, IndependentSpinOrientationGaussianIsotropic
from ..models.transformations import PrimaryMassAndMassRatioToComponentMassesTransform
from ..utils.kernel import log_planck_taper_window
from .constraints import mass_ratio_mass_sandwich, mass_sandwich
from .redshift import PowerlawRedshift
from .utils import (
    doubly_truncated_power_law_icdf,
    doubly_truncated_power_law_log_norm_constant,
    doubly_truncated_power_law_log_prob,
    ExtendedSupportTransformedDistribution,
    JointDistribution,
    ScaledMixture,
)


class PowerlawPrimaryMassRatio(Distribution):
    r"""Power law model for two-dimensional mass distribution, modelling primary mass and
    conditional mass ratio distribution.

    .. math::
        p(m_1,q\mid\alpha,\beta) = p(m_1\mid\alpha)p(q \mid m_1, \beta)

    .. math::
        \begin{align*}
            p(m_1\mid\alpha)&
            \propto m_1^{-\alpha},\qquad m_{\text{min}}\leq m_1\leq m_{\max}\\
            p(q\mid m_1,\beta)&
            \propto q^{\beta},\qquad \frac{m_{\text{min}}}{m_1}\leq q\leq 1
        \end{align*}
    """

    arg_constraints = {
        "alpha": constraints.real,
        "beta": constraints.real,
        "mmin": constraints.positive,
        "mmax": constraints.positive,
    }
    reparametrized_params = ["alpha", "beta", "mmin", "mmax"]
    pytree_data_fields = ("_support", "alpha", "beta", "mmax", "mmin")

    def __init__(
        self,
        alpha: ArrayLike,
        beta: ArrayLike,
        mmin: ArrayLike,
        mmax: ArrayLike,
        *,
        validate_args: Optional[bool] = None,
    ) -> None:
        """
        Parameters
        ----------
        alpha : ArrayLike
            Power law index for primary mass
        beta : ArrayLike
            Power law index for mass ratio
        mmin : ArrayLike
            Minimum mass
        mmax : ArrayLike
            Maximum mass
        """
        self.alpha, self.beta, self.mmin, self.mmax = promote_shapes(
            alpha, beta, mmin, mmax
        )
        batch_shape = lax.broadcast_shapes(
            jnp.shape(alpha), jnp.shape(beta), jnp.shape(mmin), jnp.shape(mmax)
        )
        self._support = mass_ratio_mass_sandwich(mmin, mmax)
        super(PowerlawPrimaryMassRatio, self).__init__(
            batch_shape=batch_shape, event_shape=(2,), validate_args=validate_args
        )

    @constraints.dependent_property(is_discrete=False, event_dim=1)
    def support(self) -> constraints.Constraint:
        return self._support

    @validate_sample
    def log_prob(self, value):
        m1, q = jnp.unstack(value, axis=-1)
        log_prob_m1 = doubly_truncated_power_law_log_prob(
            x=m1, alpha=-self.alpha, low=self.mmin, high=self.mmax
        )
        log_prob_q = jnp.where(
            jnp.less_equal(m1, self.mmin),
            -jnp.inf,
            doubly_truncated_power_law_log_prob(
                x=q, alpha=self.beta, low=self.mmin / m1, high=1.0
            ),
        )

        return log_prob_m1 + log_prob_q

    def sample(self, key, sample_shape=()):
        key_m1, key_q = jrd.split(key)
        u_m1 = jrd.uniform(key_m1, shape=sample_shape)
        u_q = jrd.uniform(key_q, shape=sample_shape)
        m1 = doubly_truncated_power_law_icdf(
            q=u_m1, alpha=-self.alpha, low=self.mmin, high=self.mmax
        )
        q = doubly_truncated_power_law_icdf(
            q=u_q, alpha=self.beta, low=jnp.divide(self.mmin, m1), high=1.0
        )
        return jnp.stack((m1, q), axis=-1)


class Wysocki2019MassModel(Distribution):
    r"""It is a double side truncated power law distribution, as described in
    equation 7 of the `Reconstructing phenomenological distributions of compact
    binaries via gravitational wave observations <https://arxiv.org/abs/1805.06442>`_.

    .. math::
        p(m_1,m_2\mid\alpha,m_{\text{min}},m_{\text{max}},M_{\text{max}})\propto
        \frac{m_1^{-\alpha}}{m_1-m_{\text{min}}}
    """

    arg_constraints = {
        "alpha_m": constraints.real,
        "mmin": constraints.positive,
        "mmax": constraints.positive,
    }
    reparametrized_params = ["alpha_m", "mmin", "mmax"]
    pytree_data_fields = ("_support", "alpha_m", "mmax", "mmin")

    def __init__(
        self,
        alpha_m: ArrayLike,
        mmin: ArrayLike,
        mmax: ArrayLike,
        *,
        validate_args: Optional[bool] = None,
    ) -> None:
        """
        Parameters
        ----------
        alpha_m : ArrayLike
            index of the power law distribution
        mmin : ArrayLike
            lower mass limit
        mmax : ArrayLike
            upper mass limit
        """
        self.alpha_m, self.mmin, self.mmax = promote_shapes(alpha_m, mmin, mmax)
        batch_shape = lax.broadcast_shapes(
            jnp.shape(alpha_m),
            jnp.shape(mmin),
            jnp.shape(mmax),
        )
        self._support = mass_sandwich(mmin, mmax)
        super(Wysocki2019MassModel, self).__init__(
            batch_shape=batch_shape, event_shape=(2,), validate_args=validate_args
        )

    @constraints.dependent_property(is_discrete=False, event_dim=1)
    def support(self) -> constraints.Constraint:
        return self._support

    @validate_sample
    def log_prob(self, value):
        m1 = value[..., 0]
        m2 = value[..., 1]
        log_prob_m1 = doubly_truncated_power_law_log_prob(
            x=m1, alpha=jnp.negative(self.alpha_m), low=self.mmin, high=self.mmax
        )
        log_prob_m2_given_m1 = uniform.logpdf(
            m2, loc=self.mmin, scale=jnp.subtract(m1, self.mmin)
        )

        return jnp.add(log_prob_m1, log_prob_m2_given_m1)

    def sample(self, key, sample_shape=()) -> Array:
        key_m1, key_m2 = jrd.split(key)
        u_m1 = jrd.uniform(key_m1, shape=sample_shape)
        m1 = doubly_truncated_power_law_icdf(
            q=u_m1, alpha=-self.alpha_m, low=self.mmin, high=self.mmax
        )
        m2 = jrd.uniform(key_m2, shape=sample_shape, minval=self.mmin, maxval=m1)
        return jnp.stack((m1, m2), axis=-1)


class SmoothedTwoComponentPrimaryMassRatio(Distribution):
    arg_constraints = {
        "alpha": constraints.real,
        "beta": constraints.real,
        "delta": constraints.positive,
        "lambda_peak": constraints.unit_interval,
        "loc": constraints.positive,
        "mmax": constraints.positive,
        "mmin": constraints.positive,
        "scale": constraints.positive,
    }
    pytree_data_fields = (
        "_logZ",
        "_m1s",
        "_support",
        "_Z_q_given_m1",
        "alpha",
        "beta",
        "delta",
        "lambda_peak",
        "loc",
        "mmax",
        "mmin",
        "scale",
    )

    def __init__(
        self,
        alpha: ArrayLike,
        beta: ArrayLike,
        delta: ArrayLike,
        lambda_peak: ArrayLike,
        loc: ArrayLike,
        mmax: ArrayLike,
        mmin: ArrayLike,
        scale: ArrayLike,
        *,
        validate_args=None,
    ) -> None:
        (
            self.alpha,
            self.beta,
            self.delta,
            self.lambda_peak,
            self.loc,
            self.mmax,
            self.mmin,
            self.scale,
        ) = promote_shapes(
            alpha,
            beta,
            delta,
            lambda_peak,
            loc,
            mmax,
            mmin,
            scale,
        )
        batch_shape = lax.broadcast_shapes(
            jnp.shape(alpha),
            jnp.shape(beta),
            jnp.shape(delta),
            jnp.shape(lambda_peak),
            jnp.shape(loc),
            jnp.shape(mmax),
            jnp.shape(mmin),
            jnp.shape(scale),
        )
        self._support = mass_ratio_mass_sandwich(mmin, mmax)
        super(SmoothedTwoComponentPrimaryMassRatio, self).__init__(
            batch_shape=batch_shape, event_shape=(2,), validate_args=validate_args
        )

        mmin = jnp.broadcast_to(mmin, batch_shape)
        mmax = jnp.broadcast_to(mmax, batch_shape)

        # Compute the normalization constant for primary mass distribution

        _m1s_delta = jnp.linspace(mmin, mmin + delta, 100)
        numerical_log_norm = jnp.trapezoid(
            jnp.exp(self._log_prob_m1_unnorm(_m1s_delta)),
            _m1s_delta,
            axis=0,
        )

        analytical_log_norm = (1 - self.lambda_peak) * jnp.exp(
            doubly_truncated_power_law_log_norm_constant(
                -self.alpha, self.mmin + self.delta, self.mmax
            )
            - doubly_truncated_power_law_log_norm_constant(
                -self.alpha, self.mmin, self.mmax
            )
        ) + self.lambda_peak * (
            special.ndtr((self.mmax - self.loc) / self.scale)
            - special.ndtr((self.mmin + self.delta - self.loc) / self.scale)
        )

        self._logZ = jnp.log(numerical_log_norm + analytical_log_norm)

        del _m1s_delta

        # Compute the normalization constant for mass ratio distribution

        self._m1s = jnp.linspace(mmin, mmax, 1000)
        _qs = jnp.linspace(0.005, 1.0, 500)
        _m1qs_grid = jnp.stack(jnp.meshgrid(self._m1s, _qs, indexing="ij"), axis=-1)

        _prob_q = jnp.exp(self._log_prob_q_unnorm(_m1qs_grid))

        self._Z_q_given_m1 = jnp.trapezoid(_prob_q, _qs, axis=1)
        del _m1qs_grid, _qs, _prob_q

    @constraints.dependent_property(is_discrete=False, event_dim=1)
    def support(self) -> constraints.Constraint:
        return self._support

    def _log_prob_m1_unnorm(self, m1: Array) -> Array:
        safe_delta = jnp.where(self.delta <= 0.0, 1.0, self.delta)
        log_smoothing_m1 = log_planck_taper_window((m1 - self.mmin) / safe_delta)
        log_norm_powerlaw = doubly_truncated_power_law_log_norm_constant(
            -self.alpha, self.mmin, self.mmax
        )
        prob_norm = norm.pdf(m1, loc=self.loc, scale=self.scale)
        log_prob_m1 = (
            jnp.log(
                (1 - self.lambda_peak)
                * jnp.power(m1, -self.alpha)
                * jnp.exp(-log_norm_powerlaw)
                + self.lambda_peak * prob_norm
            )
            + log_smoothing_m1
        )

        return jnp.where(self.delta <= 0.0, -jnp.inf, log_prob_m1)

    @validate_sample
    def _log_prob_q_unnorm(self, value: Array) -> Array:
        m1, q = jnp.unstack(value, axis=-1)
        m2 = m1 * q
        safe_delta = jnp.where(self.delta <= 0.0, 1.0, self.delta)
        log_smoothing_q = log_planck_taper_window((m2 - self.mmin) / safe_delta)
        log_prob_q = self.beta * jnp.log(q) + log_smoothing_q

        return jnp.where(self.delta <= 0.0, -jnp.inf, log_prob_q)

    @validate_sample
    def log_prob(self, value: ArrayLike) -> ArrayLike:
        m1, _ = jnp.unstack(value, axis=-1)
        log_prob_m1 = self._log_prob_m1_unnorm(m1) - self._logZ
        _Z_q = jnp.interp(m1, self._m1s, self._Z_q_given_m1, left=1.0, right=1.0)
        safe_Z_q = jnp.where(_Z_q <= 0, 1.0, _Z_q)
        log_Z_q = jnp.where(_Z_q <= 0, 0.0, jnp.log(safe_Z_q))
        log_prob_q = self._log_prob_q_unnorm(value) - log_Z_q
        return log_prob_m1 + log_prob_q


def PowerlawPeak(
    use_spin: bool = False,
    use_redshift: bool = False,
    use_tilt: bool = False,
    validate_args: Optional[bool] = None,
    **params: Array,
) -> ScaledMixture:
    # NOTE: If you change something here, please also change in
    # kokab/one_powerlaw_one_peak/ppd.py
    smoothing_model = ExtendedSupportTransformedDistribution(
        SmoothedTwoComponentPrimaryMassRatio(
            alpha=params["alpha"],
            beta=params["beta"],
            delta=params["delta"],
            lambda_peak=params["lambda_peak"],
            loc=params["loc"],
            mmax=params["mmax"],
            mmin=params["mmin"],
            scale=params["scale"],
            validate_args=validate_args,
        ),
        transforms=PrimaryMassAndMassRatioToComponentMassesTransform(),
        validate_args=validate_args,
    )

    component_distributions = [smoothing_model]

    if use_spin:
        chi_dist = Independent(
            BetaFromMeanVar(
                mean=jnp.stack([params["chi_mean"], params["chi_mean"]], axis=-1),
                variance=jnp.stack(
                    [params["chi_variance"], params["chi_variance"]], axis=-1
                ),
                validate_args=validate_args,
            ),
            reinterpreted_batch_ndims=1,
            validate_args=validate_args,
        )
        component_distributions.append(chi_dist)

    if use_tilt:
        tilt_dist = IndependentSpinOrientationGaussianIsotropic(
            zeta=params["cos_tilt_zeta"],
            scale1=params["cos_tilt_scale"],
            scale2=params["cos_tilt_scale"],
            validate_args=validate_args,
        )

        component_distributions.append(tilt_dist)

    if use_redshift:
        z_max = params["z_max"]
        kappa = params["kappa"]
        powerlaw_z = PowerlawRedshift(
            z_max=z_max, kappa=kappa, validate_args=validate_args
        )

        component_distributions.append(powerlaw_z)

    if len(component_distributions) > 1:
        component_distributions = [
            JointDistribution(*component_distributions, validate_args=validate_args)
        ]

    return ScaledMixture(
        log_scales=jnp.asarray([params["log_rate"]]),
        component_distributions=component_distributions,
        support=component_distributions[0]._support,  # type: ignore
        validate_args=validate_args,
    )



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from typing import Optional

import RIFT.lalsimutils as lalsimutils
from jax import numpy as jnp, random as jrd
from jaxtyping import Array, PRNGKeyArray
from numpyro import distributions as dist

from .utils.transformations import chirp_mass, symmetric_mass_ratio


__all__ = [
    "banana_error_m1_m2",
    "truncated_normal_error",
]


def banana_error_m1_m2(
    x: Array,
    size: int,
    key: PRNGKeyArray,
    *,
    scale_Mc: float = 1.0,
    scale_eta: float = 1.0,
) -> Array:
    r"""Add banana error to the given values. Section 3 of the `Model-independent
    inference on compact-binary observations <https://doi.org/10.1093/mnras/stw2883>`_
    discusses the banana error. It adds errors in the chirp mass and symmetric mass
    ratio and then converts back to masses.

    .. math::

        M_{c} = M_{c}^{T}
        \left[1+\beta\frac{12}{\rho}\left(r_{0}+r\right)\right]

        \eta = \eta^{T}
        \left[1+0.03\frac{12}{\rho}\left(r_{0}^{'}+r^{'}\right)\right]

    x : Array
        given values as m1 and m2
    size : int
        number of samples
    key : PRNGKeyArray
        jax random key
    scale_Mc : float
        scale of the chirp mass error, defaults to 1.0
    scale_eta : float
        scale of the symmetric mass ratio error, defaults to 1.0

    Returns
    -------
    Array
        m1 and m2 with banana
    """
    m1 = x[..., 0]
    m2 = x[..., 1]

    keys = jrd.split(key, 5)

    r0 = jrd.normal(key=keys[0])
    r0p = jrd.normal(key=keys[1])
    r = jrd.normal(key=keys[2], shape=(size,)) * scale_Mc
    rp = jrd.normal(key=keys[3], shape=(size,)) * scale_eta
    rho = 9.0 * jnp.power(jrd.uniform(key=keys[4]), -1.0 / 3.0)

    Mc_true = chirp_mass(m1=m1, m2=m2)
    eta_true = symmetric_mass_ratio(m1=m1, m2=m2)

    v_PN_param = (jnp.pi * Mc_true * 20 * lalsimutils.MsunInSec) ** (
        1.0 / 3.0
    )  # 'v' parameter
    v_PN_param_max = 0.2
    v_PN_param = jnp.min(jnp.array([v_PN_param, v_PN_param_max]))
    snr_fac = rho / 12.0
    # this ignores range due to redshift / distance, based on a low-order est
    ln_mc_error_pseudo_fisher = (
        1.5 * 0.3 * (v_PN_param / v_PN_param_max) ** (7.0) / snr_fac
    )

    beta = jnp.min(jnp.array([0.07 / snr_fac, ln_mc_error_pseudo_fisher]))

    Mc = Mc_true * (1.0 + beta * (r0 + r))
    eta = eta_true * (1.0 + 0.03 * (12.0 / rho) * (r0p + rp))

    etaV = 1.0 - 4.0 * eta
    etaV_sqrt = jnp.where(etaV >= 0, jnp.sqrt(etaV), jnp.nan)

    factor = 0.5 * Mc * jnp.power(eta, -0.6)
    m1_final = factor * (1.0 + etaV_sqrt)
    m2_final = factor * (1.0 - etaV_sqrt)

    return jnp.column_stack([m1_final, m2_final])


def truncated_normal_error(
    x: Array,
    size: int,
    key: PRNGKeyArray,
    *,
    scale: Optional[float] = None,
    low: Optional[float] = None,
    high: Optional[float] = None,
    cut_low: Optional[float] = None,
    cut_high: Optional[float] = None,
) -> Array:
    """Adds truncated normal error to the given values. The error is sampled from a
    truncated normal distribution with the given parameters. The function will resample
    until all values are within the allowed range.

    .. note::

        If :code:`cut_low` and :code:`cut_high` are both None, the function will return
        the sampled values without any truncation.

    .. note::

        if :code:`low` or :code:`high` is None, the :code:`TruncatedNormal` distribution
        will not be truncated at those bounds.

    Parameters
    ----------
    x : Array
        Given values to which the error will be added.
    size : int
        Number of samples to generate.
    key : PRNGKeyArray
        JAX random key for sampling.
    scale : Optional[float], optional
        Scale parameter for the truncated normal distribution, by default None
    low : Optional[float], optional
        Lower bound for the truncated normal distribution, by default None
    high : Optional[float], optional
        Upper bound for the truncated normal distribution, by default None
    cut_low : Optional[float], optional
        Lower bound for the final values, by default None
    cut_high : Optional[float], optional
        Upper bound for the final values, by default None

    Returns
    -------
    Array
        Array of values with added truncated normal error.

    Raises
    ------
    ValueError
        If the scale parameter is not provided.
    """
    if scale is None:
        raise ValueError("Scale parameter is required.")

    err_dist = dist.TruncatedNormal(loc=x, scale=scale, low=low, high=high)

    # Initial sampling from the truncated normal distribution.
    err_x = err_dist.sample(key=key, sample_shape=(size,))

    if cut_low is None and cut_high is None:
        return err_x

    # Resample until all values are within the allowed range
    while True:
        mask = jnp.zeros_like(err_x, dtype=bool)
        if cut_low is not None:
            mask = mask | (err_x < cut_low)
        if cut_high is not None:
            mask = mask | (err_x > cut_high)

        # If no values are outside the allowed range, break the loop
        # This is important to avoid infinite loops
        if not jnp.any(mask).item():
            break

        # Split the key for a new random seed
        key, _ = jrd.split(key)
        num_invalid = int(jnp.sum(mask))
        new_samples = err_dist.sample(key=key, sample_shape=(num_invalid,))
        invalid_indices = jnp.where(mask)[0]
        err_x = err_x.at[invalid_indices].set(new_samples)

    return err_x



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from collections.abc import Sequence
from typing import Dict, Optional, Union

import h5py
import jax
import numpy as np
from jaxtyping import Array

from gwkokab.constants import SECONDS_PER_YEAR

from ..utils.tools import error_if, warn_if
from ._abc import VolumeTimeSensitivityInterface


class SyntheticInjectionVolumeTimeSensitivity(VolumeTimeSensitivityInterface):
    injections: Array
    """Array of real injections of shape (n_injections, n_features)."""
    sampling_prob: Array
    """Array of sampling probabilities of shape (n_injections,)."""
    analysis_time_years: float
    """Analysis time in years."""
    total_injections: int
    """Total number of injections."""

    def __init__(
        self, parameters: Sequence[str], filename: str, batch_size: Optional[int] = None
    ) -> None:
        """Convenience class for loading a neural vt.

        Parameters
        ----------
        parameters : Sequence[str]
            The names of the parameters that the model expects.
        filename : str
            The filename of the neural vt.
        batch_size: Optional[int]
            The batch size of the neural vt.
        spin_case : Literal[&quot;aligned_spin&quot;, &quot;full_spin&quot;]
            The spin case of the injections. Default is 'aligned_spin'.
        """
        error_if(not parameters, msg="parameters sequence cannot be empty")
        error_if(
            not isinstance(parameters, Sequence),
            msg=f"parameters must be a Sequence, got {type(parameters)}",
        )
        error_if(
            not all(isinstance(p, str) for p in parameters),
            msg="all parameters must be strings",
        )
        warn_if(
            batch_size is not None,
            msg="batch_size is not used for injection based VTs",
        )

        with h5py.File(filename, "r") as f:
            self.analysis_time_years = (
                float(f.attrs["analysis_time_s"]) / SECONDS_PER_YEAR
            )
            injs = [f[p][:] for p in parameters]
            self.injections = jax.device_put(np.stack(injs, axis=-1), may_alias=True)
            self.sampling_prob = jax.device_put(f["sampling_pdf"][:], may_alias=True)
            self.total_injections = self.sampling_prob.shape[0]
            param_ranges: Dict[str, Union[int, float]] = {}

            for i, p in enumerate(parameters):
                try:
                    minimum = f.attrs[p + "_min"]
                except KeyError:
                    minimum = np.min(self.injections[:, i])
                try:
                    maximum = f.attrs[p + "_max"]
                except KeyError:
                    maximum = np.max(self.injections[:, i])

                param_ranges[p + "_min"] = minimum
                param_ranges[p + "_max"] = maximum
            self.parameter_ranges = param_ranges

    def get_logVT(self):
        raise NotImplementedError("Injection based VTs do not have a logVT method.")

    def get_mapped_logVT(self):
        raise NotImplementedError("Injection based VTs do not have a logVT method.")



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import argparse
import glob
import os

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt


def make_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Command line interface for plotting chains.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        epilog="This script plots chains from .dat files.",
    )
    parser.add_argument(
        "-d",
        "--data-regex",
        help="regex pattern for the data files. Only .dat files are supported.",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-o",
        "--output",
        help="output file path",
        required=True,
        type=argparse.FileType("w"),
    )
    parser.add_argument(
        "--band",
        help="plot the band instead of the chains",
        action="store_true",
    )
    parser.add_argument(
        "--band-color",
        help="color of the band",
        default="#1f77b4",
    )
    parser.add_argument(
        "-l",
        "--labels",
        nargs="+",
        help="labels for the chains",
        required=True,
    )
    parser.add_argument(
        "-a",
        "--alpha",
        help="transparency of the chains",
        default=0.3,
        type=float,
    )
    parser.add_argument(
        "-t",
        "--title",
        help="title for the plot",
        type=str,
    )
    parser.add_argument(
        "-width",
        "--width",
        help="width of the plot in inches",
        default=20,
        type=int,
    )
    parser.add_argument(
        "-height",
        "--height",
        help="height of the plot in inches",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-xs",
        "--x-scale",
        help="scale of the x-axis",
        default="linear",
        type=str,
    )
    parser.add_argument(
        "-ys",
        "--y-scale",
        help="scale of the y-axis",
        default="linear",
        type=str,
    )
    parser.add_argument(
        "--use-latex",
        help="use LaTeX for rendering text",
        action="store_true",
    )
    parser.add_argument(
        "--font-family",
        help="font family to use",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--dpi",
        help="dots per inch to save file",
        type=int,
        default=100,
    )

    return parser


def main() -> None:
    parser = make_parser()
    args = parser.parse_args()

    plt.rcParams.update({"text.usetex": args.use_latex})
    if args.font_family is not None:
        plt.rcParams.update({"font.family": args.font_family})

    files = glob.glob(args.data_regex)

    n_dim = pd.read_csv(files[0], delimiter=" ", skiprows=1).to_numpy().shape[1]

    if args.height is None:
        figsize = (args.width, n_dim * 2.5)
    else:
        figsize = (args.width, args.height)

    band = args.band
    band_color = args.band_color
    _, ax = plt.subplots(n_dim, 1, figsize=figsize, sharex=True)
    if n_dim == 1:
        for file in files:
            data = pd.read_csv(file, delimiter=" ", skiprows=1).to_numpy()
            n_points = data.shape[0]
            if not band:
                ax.plot(data, alpha=args.alpha)
            else:
                quantiles = np.quantile(data, [0.05, 0.25, 0.75, 0.95], axis=-1)
                mean_values = np.mean(data, axis=-1)
                ax.fill_between(
                    range(n_points),
                    quantiles[0],
                    quantiles[-1],
                    alpha=0.2,
                    color=band_color,
                )
                ax.fill_between(
                    range(n_points),
                    quantiles[1],
                    quantiles[-2],
                    alpha=0.5,
                    color=band_color,
                )
                ax.plot(
                    range(n_points),
                    mean_values,
                    color=band_color,
                    alpha=1.0,
                )
            ax.set_ylabel(args.labels[0])
            plt.tick_params(
                axis="both",
                which="both",
                labelleft=True,
                labelright=True,
                labeltop=True,
                labelbottom=True,
            )
            plt.grid(visible=True, which="both", axis="both", alpha=0.5)
    else:
        data = np.stack(
            [pd.read_csv(file, delimiter=" ", skiprows=1).to_numpy() for file in files],
            axis=-1,
        )
        n_points = data.shape[0]
        if band:
            quantiles = np.quantile(data, [0.05, 0.25, 0.75, 0.95], axis=-1)
            mean_values = np.mean(data, axis=-1)

        for i in range(n_dim):
            ax[i].set_ylabel(args.labels[i])
            ax[i].tick_params(
                axis="both",
                which="both",
                labelleft=True,
                labelright=True,
                labeltop=i == 0,
                labelbottom=i == n_dim - 1,
            )
            ax[i].grid(visible=True, which="both", axis="both", alpha=0.5)
            if not band:
                data_ = data[:, i, :]
                ax[i].plot(data_, alpha=args.alpha)
            else:
                quantiles_ = quantiles[:, :, i]
                mean_values_ = mean_values[:, i]
                ax[i].fill_between(
                    range(n_points),
                    quantiles_[0],
                    quantiles_[-1],
                    alpha=0.2,
                    color=band_color,
                )
                ax[i].fill_between(
                    range(n_points),
                    quantiles_[1],
                    quantiles_[-2],
                    alpha=0.5,
                    color=band_color,
                )
                ax[i].plot(
                    range(n_points),
                    mean_values_,
                    color=band_color,
                    alpha=1.0,
                )
    if args.title:
        plt.suptitle(args.title)
    plt.xscale(args.x_scale)
    plt.yscale(args.y_scale)
    plt.tight_layout()
    # Determine output file type and save accordingly
    output_ext = os.path.splitext(args.output.name)[1].lower()
    plt_savefig_kwargs = dict()
    if output_ext == ".png":
        plt_savefig_kwargs["dpi"] = args.dpi
    plt.savefig(args.output.name, bbox_inches="tight", **plt_savefig_kwargs)
    plt.close("all")



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import functools as ft
import gc
import os
from collections.abc import Callable
from typing import Any, Dict, List, Optional

import equinox as eqx
import jax
import numpy as np
import tqdm
from flowMC.resource.base import Resource
from flowMC.resource.buffers import Buffer
from flowMC.resource.local_kernel.MALA import MALA
from flowMC.resource.logPDF import LogPDF
from flowMC.resource.nf_model.NF_proposal import NFProposal
from flowMC.resource.nf_model.rqSpline import MaskedCouplingRQSpline
from flowMC.resource.optimizer import Optimizer
from flowMC.resource.states import State
from flowMC.resource_strategy_bundle.base import ResourceStrategyBundle
from flowMC.strategy.base import Strategy
from flowMC.strategy.lambda_function import Lambda
from flowMC.strategy.take_steps import TakeGroupSteps, TakeSerialSteps
from flowMC.strategy.train_model import TrainModel
from flowMC.strategy.update_state import UpdateState
from jax import numpy as jnp, random as jrd
from jaxtyping import Array, Float, PRNGKeyArray
from loguru import logger

from gwkokab.models.utils import JointDistribution
from gwkokab.utils.tools import error_if
from kokab.utils.common import read_json
from kokab.utils.guru import Guru, guru_arg_parser
from kokab.utils.literals import INFERENCE_DIRECTORY


# WARNING: do not change anything in this class


# Copyright (c) 2022 Kaze Wong & contributor
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
class RQSpline_MALA_Bundle(ResourceStrategyBundle):
    """A bundle that uses a Rational Quadratic Spline as a normalizing flow model and
    the Metropolis Adjusted Langevin Algorithm as a local sampler.

    This is the base algorithm described in
    https://www.pnas.org/doi/full/10.1073/pnas.2109420119
    """

    def __repr__(self):
        return "RQSpline_MALA Bundle"

    def __init__(
        self,
        rng_key: PRNGKeyArray,
        n_chains: int,
        n_dims: int,
        logpdf: Callable[[Float[Array, " n_dim"], dict], Float],
        n_local_steps: int,
        n_global_steps: int,
        n_training_loops: int,
        n_production_loops: int,
        n_epochs: int,
        mala_step_size: float = 1e-1,
        chain_batch_size: int = 0,
        rq_spline_hidden_units: list[int] = [32, 32],
        rq_spline_n_bins: int = 8,
        rq_spline_n_layers: int = 4,
        rq_spline_range: tuple[float, float] = (-10.0, 10.0),
        learning_rate: float = 1e-3,
        batch_size: int = 10000,
        n_max_examples: int = 10000,
        history_window: int = 100,
        local_thinning: int = 1,
        global_thinning: int = 1,
        n_NFproposal_batch_size: int = 10000,
        verbose: bool = False,
    ):
        n_training_steps = (
            n_local_steps // local_thinning * n_training_loops
            + n_global_steps // global_thinning * n_training_loops
        )
        n_production_steps = (
            n_local_steps // local_thinning * n_production_loops
            + n_global_steps // global_thinning * n_production_loops
        )
        n_total_epochs = n_training_loops * n_epochs

        positions_training = Buffer(
            "positions_training", (n_chains, n_training_steps, n_dims), 1
        )
        log_prob_training = Buffer("log_prob_training", (n_chains, n_training_steps), 1)
        local_accs_training = Buffer(
            "local_accs_training", (n_chains, n_training_steps), 1
        )
        global_accs_training = Buffer(
            "global_accs_training", (n_chains, n_training_steps), 1
        )
        loss_buffer = Buffer("loss_buffer", (n_total_epochs,), 0)

        position_production = Buffer(
            "positions_production", (n_chains, n_production_steps, n_dims), 1
        )
        log_prob_production = Buffer(
            "log_prob_production", (n_chains, n_production_steps), 1
        )
        local_accs_production = Buffer(
            "local_accs_production", (n_chains, n_production_steps), 1
        )
        global_accs_production = Buffer(
            "global_accs_production", (n_chains, n_production_steps), 1
        )

        local_sampler = MALA(step_size=mala_step_size)
        rng_key, subkey = jax.random.split(rng_key)
        model = MaskedCouplingRQSpline(
            n_dims,
            rq_spline_n_layers,
            rq_spline_hidden_units,
            rq_spline_n_bins,
            subkey,
            rq_spline_range,
        )
        global_sampler = NFProposal(
            model, n_NFproposal_batch_size=n_NFproposal_batch_size
        )
        optimizer = Optimizer(model=model, learning_rate=learning_rate)
        logpdf = LogPDF(logpdf, n_dims=n_dims)

        sampler_state = State(
            {
                "target_positions": "positions_training",
                "target_log_prob": "log_prob_training",
                "target_local_accs": "local_accs_training",
                "target_global_accs": "global_accs_training",
                "training": True,
            },
            name="sampler_state",
        )

        self.resources = {
            "logpdf": logpdf,
            "positions_training": positions_training,
            "log_prob_training": log_prob_training,
            "local_accs_training": local_accs_training,
            "global_accs_training": global_accs_training,
            "loss_buffer": loss_buffer,
            "positions_production": position_production,
            "log_prob_production": log_prob_production,
            "local_accs_production": local_accs_production,
            "global_accs_production": global_accs_production,
            "local_sampler": local_sampler,
            "global_sampler": global_sampler,
            "model": model,
            "optimizer": optimizer,
            "sampler_state": sampler_state,
        }

        local_stepper = TakeSerialSteps(
            "logpdf",
            "local_sampler",
            "sampler_state",
            ["target_positions", "target_log_prob", "target_local_accs"],
            n_local_steps,
            thinning=local_thinning,
            chain_batch_size=chain_batch_size,
            verbose=verbose,
        )

        global_stepper = TakeGroupSteps(
            "logpdf",
            "global_sampler",
            "sampler_state",
            ["target_positions", "target_log_prob", "target_global_accs"],
            n_global_steps,
            thinning=global_thinning,
            chain_batch_size=chain_batch_size,
            verbose=verbose,
        )

        model_trainer = TrainModel(
            "model",
            "positions_training",
            "optimizer",
            loss_buffer_name="loss_buffer",
            n_epochs=n_epochs,
            batch_size=batch_size,
            n_max_examples=n_max_examples,
            history_window=history_window,
            verbose=verbose,
        )

        update_state = UpdateState(
            "sampler_state",
            [
                "target_positions",
                "target_log_prob",
                "target_local_accs",
                "target_global_accs",
                "training",
            ],
            [
                "positions_production",
                "log_prob_production",
                "local_accs_production",
                "global_accs_production",
                False,
            ],
        )

        def reset_steppers(
            rng_key: PRNGKeyArray,
            resources: dict[str, Resource],
            initial_position: Float[Array, "n_chains n_dim"],
            data: dict,
        ) -> tuple[
            PRNGKeyArray,
            dict[str, Resource],
            Float[Array, "n_chains n_dim"],
        ]:
            """Reset the steppers to the initial position."""
            local_stepper.set_current_position(0)
            global_stepper.set_current_position(0)
            return rng_key, resources, initial_position

        reset_steppers_lambda = Lambda(
            lambda rng_key, resources, initial_position, data: reset_steppers(
                rng_key, resources, initial_position, data
            )
        )

        update_global_step = Lambda(
            lambda rng_key,
            resources,
            initial_position,
            data: global_stepper.set_current_position(local_stepper.current_position)
        )
        update_local_step = Lambda(
            lambda rng_key,
            resources,
            initial_position,
            data: local_stepper.set_current_position(global_stepper.current_position)
        )

        def update_model(
            rng_key: PRNGKeyArray,
            resources: dict[str, Resource],
            initial_position: Float[Array, "n_chains n_dim"],
            data: dict,
        ) -> tuple[
            PRNGKeyArray,
            dict[str, Resource],
            Float[Array, "n_chains n_dim"],
        ]:
            """Update the model."""
            model = resources["model"]
            resources["global_sampler"] = eqx.tree_at(
                lambda x: x.model,
                resources["global_sampler"],
                model,
            )
            return rng_key, resources, initial_position

        update_model_lambda = Lambda(
            lambda rng_key, resources, initial_position, data: update_model(
                rng_key, resources, initial_position, data
            )
        )

        self.strategies = {
            "local_stepper": local_stepper,
            "global_stepper": global_stepper,
            "model_trainer": model_trainer,
            "update_state": update_state,
            "update_global_step": update_global_step,
            "update_local_step": update_local_step,
            "reset_steppers": reset_steppers_lambda,
            "update_model": update_model_lambda,
        }

        training_phase = [
            "local_stepper",
            "update_global_step",
            "model_trainer",
            "update_model",
            "global_stepper",
            "update_local_step",
        ]
        production_phase = [
            "local_stepper",
            "update_global_step",
            "global_stepper",
            "update_local_step",
        ]
        strategy_order = []
        for _ in range(n_training_loops):
            strategy_order.extend(training_phase)

        strategy_order.append("reset_steppers")
        strategy_order.append("update_state")
        for _ in range(n_production_loops):
            strategy_order.extend(production_phase)

        self.strategy_order = strategy_order


# WARNING: do not change anything in this class


# Copyright (c) 2022 Kaze Wong & contributor
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
class Sampler:
    """Top level API that the users primarily interact with.

    Args:
        n_dim (int): Dimension of the parameter space.
        n_chains (int): Number of chains to sample.
        rng_key (PRNGKeyArray): Jax PRNGKey.
        logpdf (Callable[[Float[Array, "n_dim"], dict], Float):
            Log probability function.
        resources (dict[str, Resource]): Resources to be used by the sampler.
        strategies (dict[str, Strategy]): Strategies to be used by the sampler.
        verbose (bool): Whether to print out progress. Defaults to False.
        logging (bool): Whether to log the progress. Defaults to True.
        outdir (str): Directory to save the logs. Defaults to "./outdir/".
    """

    # Essential parameters
    n_dim: int
    n_chains: int
    rng_key: PRNGKeyArray
    resources: dict[str, Resource]
    strategies: dict[str, Strategy]
    strategy_order: Optional[list[str]]

    # Logging hyperparameters
    verbose: bool = False
    logging: bool = True
    outdir: str = "./outdir/"

    def __init__(
        self,
        n_dim: int,
        n_chains: int,
        rng_key: PRNGKeyArray,
        resources: None | dict[str, Resource] = None,
        strategies: None | dict[str, Strategy] = None,
        strategy_order: None | list[str] = None,
        resource_strategy_bundles: None | ResourceStrategyBundle = None,
        **kwargs,
    ):
        # Copying input into the model

        self.n_dim = n_dim
        self.n_chains = n_chains
        self.rng_key = rng_key

        if resources is not None and strategies is not None:
            print(
                "Resources and strategies provided. Ignoring resource strategy bundles."
            )
            self.resources = resources
            self.strategies = strategies
            self.strategy_order = strategy_order

        else:
            print(
                "Resources or strategies not provided. Using resource strategy bundles."
            )
            if resource_strategy_bundles is None:
                raise ValueError(
                    "Resource strategy bundles not provided."
                    "Please provide either resources and strategies or resource strategy bundles."
                )
            self.resources = resource_strategy_bundles.resources
            self.strategies = resource_strategy_bundles.strategies
            self.strategy_order = resource_strategy_bundles.strategy_order

        # Set and override any given hyperparameters
        class_keys = list(self.__class__.__dict__.keys())
        for key, value in kwargs.items():
            if key in class_keys:
                if not key.startswith("__"):
                    setattr(self, key, value)

    def sample(self, initial_position: Float[Array, "n_chains n_dim"], data: dict):
        """Sample from the posterior using the local sampler.

        Args:
            initial_position (Device Array): Initial position.
            data (dict): Data to be used by the likelihood functions
        """

        initial_position = jnp.atleast_2d(initial_position)  # type: ignore
        rng_key = self.rng_key
        last_step = initial_position
        assert isinstance(self.strategy_order, list)
        for strategy in self.strategy_order:
            if strategy not in self.strategies:
                raise ValueError(
                    f"Invalid strategy name '{strategy}' provided. "
                    f"Available strategies are: {list(self.strategies.keys())}."
                )

        n_total_strategy = len(self.strategy_order)
        n_global_steps = self.strategy_order.count("model_trainer")
        n_local_steps = (n_total_strategy - 6 * n_global_steps - 2) // 4

        with tqdm.tqdm(range(n_global_steps), total=n_global_steps) as pbar:
            pbar.set_description("Global Tuning")
            for i in pbar:
                for strategy in self.strategy_order[6 * i : 6 * (i + 1)]:
                    (
                        rng_key,
                        self.resources,
                        last_step,
                    ) = self.strategies[strategy](
                        rng_key, self.resources, last_step, data
                    )

        print("Resetting steppers and updating state")
        for strategy in self.strategy_order[
            6 * n_global_steps : 6 * n_global_steps + 2
        ]:
            (
                rng_key,
                self.resources,
                last_step,
            ) = self.strategies[strategy](rng_key, self.resources, last_step, data)

        with tqdm.tqdm(range(n_local_steps), total=n_local_steps) as pbar:
            pbar.set_description("Global Sampling")
            offset = 6 * n_global_steps + 2
            for i in pbar:
                for strategy in self.strategy_order[
                    offset + 4 * i : offset + 4 * (i + 1)
                ]:
                    (
                        rng_key,
                        self.resources,
                        last_step,
                    ) = self.strategies[strategy](
                        rng_key, self.resources, last_step, data
                    )

    # TODO: Implement quick access and summary functions that operates on buffer

    def serialize(self):
        """Serialize the sampler object."""
        raise NotImplementedError

    def deserialize(self):
        """Deserialize the sampler object."""
        raise NotImplementedError


def _same_length_arrays(length: int, *arrays: np.ndarray) -> tuple[np.ndarray, ...]:
    """This function pads the arrays with None to make them the same length.

    Parameters
    ----------
    length : int
        The length of the arrays.
    arrays : np.ndarray
        The arrays to pad.

    Returns
    -------
    tuple[np.ndarray, ...]
        The padded arrays.
    """
    padded_arrays = []
    for array in arrays:
        padded_array = np.empty((length,))
        padded_array[..., : array.shape[0]] = array
        padded_array[..., array.shape[0] :] = None
        padded_arrays.append(padded_array)
    return tuple(padded_arrays)


def _save_data_from_sampler(
    sampler: Sampler,
    *,
    rng_key: PRNGKeyArray,
    out_dir: str,
    labels: Optional[list[str]] = None,
    n_samples: int = 5000,
    logpdf: Optional[Callable] = None,
) -> None:
    """This functions saves the data from a sampler to disk. The data saved includes the
    samples from the flow, the chains from the training and production phases, the log
    probabilities, the global and local acceptance rates, and the loss values.

    Parameters
    ----------
    sampler : Sampler
        The sampler object.
    out_dir : str
        The output directory.
    labels : Optional[list[str]], optional
        list of labels for the samples, by default None
    n_samples : int, optional
        number of samples to draw from the flow, by default 5000
    logpdf : Callable, optional
        The log probability function.
    """
    logger.info("Saving data from sampler to directory: {out_dir}", out_dir=out_dir)
    if labels is None:
        labels = [f"x{i}" for i in range(sampler.n_dim)]
    if logpdf is None:
        raise ValueError("logpdf must be provided")

    os.makedirs(out_dir, exist_ok=True)

    header = " ".join(labels)

    sampler_resources = sampler.resources

    train_chains = np.array(sampler_resources["positions_training"].data)
    train_global_accs = np.array(sampler_resources["global_accs_training"].data)
    train_local_accs = np.array(sampler_resources["local_accs_training"].data)
    train_loss_vals = np.array(sampler_resources["loss_buffer"].data)
    train_log_prob = np.array(sampler_resources["log_prob_training"].data)

    prod_chains = np.array(sampler_resources["positions_production"].data)
    prod_global_accs = np.array(sampler_resources["global_accs_production"].data)
    prod_local_accs = np.array(sampler_resources["local_accs_production"].data)
    prod_log_prob = np.array(sampler_resources["log_prob_production"].data)

    n_chains = sampler.n_chains

    np.savetxt(
        rf"{out_dir}/global_accs.dat",
        np.column_stack(
            _same_length_arrays(
                max(train_global_accs.shape[1], prod_global_accs.shape[1]),
                train_global_accs.mean(0),
                prod_global_accs.mean(0),
            )
        ),
        header="train prod",
        comments="#",
    )
    logger.info(
        "Global acceptance rates saved to {out_dir}/global_accs.dat", out_dir=out_dir
    )

    np.savetxt(
        rf"{out_dir}/local_accs.dat",
        np.column_stack(
            _same_length_arrays(
                max(train_local_accs.shape[1], prod_local_accs.shape[1]),
                train_local_accs.mean(0),
                prod_local_accs.mean(0),
            )
        ),
        header="train prod",
        comments="#",
    )
    logger.info(
        "Local acceptance rates saved to {out_dir}/local_accs.dat", out_dir=out_dir
    )

    np.savetxt(rf"{out_dir}/loss.dat", train_loss_vals.reshape(-1), header="loss")
    logger.info("Loss values saved to {out_dir}/loss.dat", out_dir=out_dir)

    for n_chain in tqdm.trange(
        n_chains, total=n_chains, unit="chain", desc="Saving data from chains"
    ):
        np.savetxt(
            rf"{out_dir}/global_accs_{n_chain}.dat",
            np.column_stack(
                _same_length_arrays(
                    max(
                        train_global_accs[n_chain, :].shape[0],
                        prod_global_accs[n_chain, :].shape[0],
                    ),
                    train_global_accs[n_chain, :],
                    prod_global_accs[n_chain, :],
                )
            ),
            header="train prod",
            comments="#",
        )

        np.savetxt(
            rf"{out_dir}/local_accs_{n_chain}.dat",
            np.column_stack(
                _same_length_arrays(
                    max(
                        train_local_accs[n_chain, :].shape[0],
                        prod_local_accs[n_chain, :].shape[0],
                    ),
                    train_local_accs[n_chain, :],
                    prod_local_accs[n_chain, :],
                )
            ),
            header="train prod",
            comments="#",
        )

        np.savetxt(
            rf"{out_dir}/train_chains_{n_chain}.dat",
            train_chains[n_chain, :, :],
            header=header,
        )
        np.savetxt(
            rf"{out_dir}/prod_chains_{n_chain}.dat",
            prod_chains[n_chain, :, :],
            header=header,
        )
        np.savetxt(
            rf"{out_dir}/log_prob_{n_chain}.dat",
            np.column_stack(
                _same_length_arrays(
                    max(
                        train_log_prob[n_chain].shape[0],
                        prod_log_prob[n_chain].shape[0],
                    ),
                    train_log_prob[n_chain],
                    prod_log_prob[n_chain],
                )
            ),
            header="train prod",
            comments="#",
        )

    gc.collect()
    logger.debug(
        "Garbage collection completed after saving chains and log probabilities."
    )

    nf_model = sampler_resources["model"]
    _, subkey = jrd.split(rng_key)
    unweighted_samples = np.asarray(
        jax.block_until_ready(nf_model.sample(n_samples=n_samples, rng_key=subkey))
    )
    np.savetxt(
        rf"{out_dir}/nf_samples_unweighted.dat", unweighted_samples, header=header
    )
    logger.info(
        "Unweighted samples saved to {out_dir}/nf_samples_unweighted.dat",
        out_dir=out_dir,
    )

    gc.collect()
    logger.debug("Garbage collection completed after unweighted samples.")

    logpdf_val = jax.block_until_ready(
        jax.lax.map(lambda s: logpdf(s), unweighted_samples)
    )
    logger.debug(
        "Nan count in logpdf values: {nan_count}", nan_count=jnp.isnan(logpdf_val).sum()
    )
    logger.debug(
        "Neginf count in logpdf values: {inf_count}",
        inf_count=jnp.isneginf(logpdf_val).sum(),
    )
    logger.debug(
        "Posinf count in logpdf values: {inf_count}",
        inf_count=jnp.isposinf(logpdf_val).sum(),
    )

    nf_model_log_prob_val = jax.block_until_ready(
        jax.lax.map(nf_model.log_prob, unweighted_samples)
    )
    logger.debug(
        "Nan count in nf_model_log_prob values: {nan_count}",
        nan_count=jnp.isnan(nf_model_log_prob_val).sum(),
    )
    logger.debug(
        "Neginf count in nf_model_log_prob values: {inf_count}",
        inf_count=jnp.isneginf(nf_model_log_prob_val).sum(),
    )
    logger.debug(
        "Posinf count in nf_model_log_prob values: {inf_count}",
        inf_count=jnp.isposinf(nf_model_log_prob_val).sum(),
    )

    weights = jax.nn.softmax(logpdf_val - nf_model_log_prob_val)
    logger.debug(
        "Nan count in weights values: {nan_count}", nan_count=jnp.isnan(weights).sum()
    )
    logger.debug(
        "Neginf count in weights values: {inf_count}",
        inf_count=jnp.isneginf(weights).sum(),
    )
    logger.debug(
        "Posinf count in weights values: {inf_count}",
        inf_count=jnp.isposinf(weights).sum(),
    )

    weights = np.asarray(weights)  # type: ignore
    ess = int(1.0 / np.sum(np.square(weights)))
    logger.debug("Effective sample size is {} out of {}".format(ess, n_samples))
    if ess == 0:
        raise ValueError("Effective sample size is zero, cannot proceed with sampling.")
    weighted_samples = unweighted_samples[
        np.random.choice(n_samples, size=ess, p=weights)
    ]
    np.savetxt(rf"{out_dir}/nf_samples_weighted.dat", weighted_samples, header=header)
    logger.info(
        "Weighted samples saved to {out_dir}/nf_samples_weighted.dat", out_dir=out_dir
    )

    gc.collect()
    logger.debug("Garbage collection completed after weighted samples.")


class FlowMCBased(Guru):
    def driver(
        self,
        *,
        logpdf: Callable[[Array, Dict[str, Any]], Array],
        priors: JointDistribution,
        data: Dict[str, Any],
        labels: List[str],
    ) -> None:
        sampler_config = read_json(self.sampler_settings_filename)

        RQSpline_MALA_Bundle_value: dict = sampler_config.pop(
            "RQSpline_MALA_Bundle", {}
        )

        batch_size = RQSpline_MALA_Bundle_value["batch_size"]
        chain_batch_size = RQSpline_MALA_Bundle_value["chain_batch_size"]
        global_thinning = RQSpline_MALA_Bundle_value["global_thinning"]
        learning_rate = RQSpline_MALA_Bundle_value["learning_rate"]
        local_thinning = RQSpline_MALA_Bundle_value["local_thinning"]
        mala_step_size = RQSpline_MALA_Bundle_value["mala_step_size"]
        n_chains = RQSpline_MALA_Bundle_value["n_chains"]
        n_epochs = RQSpline_MALA_Bundle_value["n_epochs"]
        n_global_steps = RQSpline_MALA_Bundle_value["n_global_steps"]
        n_local_steps = RQSpline_MALA_Bundle_value["n_local_steps"]
        n_max_examples = RQSpline_MALA_Bundle_value["n_max_examples"]
        history_window = RQSpline_MALA_Bundle_value.get("history_window", 100)
        n_NFproposal_batch_size = RQSpline_MALA_Bundle_value["n_NFproposal_batch_size"]
        n_production_loops = RQSpline_MALA_Bundle_value["n_production_loops"]
        n_training_loops = RQSpline_MALA_Bundle_value["n_training_loops"]
        rq_spline_hidden_units = RQSpline_MALA_Bundle_value["rq_spline_hidden_units"]
        rq_spline_n_bins = RQSpline_MALA_Bundle_value["rq_spline_n_bins"]
        rq_spline_n_layers = RQSpline_MALA_Bundle_value["rq_spline_n_layers"]
        rq_spline_range = RQSpline_MALA_Bundle_value.get(
            "rq_spline_range", (-10.0, 10.0)
        )
        rq_spline_range = tuple(rq_spline_range)
        verbose = RQSpline_MALA_Bundle_value["verbose"]

        logger.debug("Validation for Sampler parameters starting")

        for var, var_name, expected_type in (
            (batch_size, "batch_size", int),
            (chain_batch_size, "chain_batch_size", int),
            (global_thinning, "global_thinning", int),
            (learning_rate, "learning_rate", (int, float)),
            (local_thinning, "local_thinning", int),
            (mala_step_size, "mala_step_size", (int, float)),
            (n_chains, "n_chains", int),
            (n_epochs, "n_epochs", int),
            (n_global_steps, "n_global_steps", int),
            (n_local_steps, "n_local_steps", int),
            (n_max_examples, "n_max_examples", int),
            (history_window, "history_window", int),
            (n_NFproposal_batch_size, "n_NFproposal_batch_size", int),
            (n_production_loops, "n_production_loops", int),
            (n_training_loops, "n_training_loops", int),
            (rq_spline_hidden_units, "rq_spline_hidden_units", list),
            (rq_spline_n_bins, "rq_spline_n_bins", int),
            (rq_spline_n_layers, "rq_spline_n_layers", int),
            (rq_spline_range, "rq_spline_range", tuple),
            (verbose, "verbose", bool),
        ):
            error_if(
                not isinstance(var, expected_type),
                err=TypeError,
                msg=f"expected a {expected_type}, got {type(var)} for {var_name}",
            )
            if expected_type is int:
                error_if(
                    var < 1 and var_name != "chain_batch_size",
                    err=ValueError,
                    msg=f"expected a positive integer, got {var} for {var_name}",
                )
            logger.debug(f"{var_name}: {var}")
        error_if(
            not all(isinstance(x, int) and x > 0 for x in rq_spline_hidden_units),
            msg=f"expected a list of positive integers, got {rq_spline_hidden_units} for rq_spline_hidden_units",
        )

        initial_position = priors.sample(self.rng_key, (n_chains,))

        n_dims = initial_position.shape[1]

        bundle = RQSpline_MALA_Bundle(
            rng_key=self.rng_key,
            n_chains=n_chains,
            n_dims=n_dims,
            logpdf=logpdf,
            n_local_steps=n_local_steps,
            n_global_steps=n_global_steps,
            n_training_loops=n_training_loops,
            n_production_loops=n_production_loops,
            n_epochs=n_epochs,
            mala_step_size=mala_step_size,
            chain_batch_size=chain_batch_size,
            rq_spline_hidden_units=rq_spline_hidden_units,
            rq_spline_n_bins=rq_spline_n_bins,
            rq_spline_n_layers=rq_spline_n_layers,
            rq_spline_range=rq_spline_range,
            learning_rate=learning_rate,
            batch_size=batch_size,
            n_max_examples=n_max_examples,
            history_window=history_window,
            local_thinning=local_thinning,
            global_thinning=global_thinning,
            n_NFproposal_batch_size=n_NFproposal_batch_size,
            verbose=verbose,
        )
        logger.info("RQSpline_MALA_Bundle created successfully.")

        sampler = Sampler(
            n_dims,
            n_chains,
            self.rng_key,
            resource_strategy_bundles=bundle,
        )

        logger.debug("Sampler initialized, starting sampling.")

        if self.debug_nans:
            with jax.debug_nans(True):
                sampler.sample(initial_position, data)
        elif self.profile_memory:
            sampler.sample(initial_position, data)
            import datetime

            time = datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
            filename = f"{self.analysis_name}_memory_{time}.prof"
            jax.profiler.save_device_memory_profile(filename)
            logger.debug("Memory profile saved as {filename}", filename=filename)
        elif self.check_leaks:
            with jax.checking_leaks():
                sampler.sample(initial_position, data)
        else:
            sampler.sample(initial_position, data)

        _save_data_from_sampler(
            sampler,
            rng_key=self.rng_key,
            logpdf=ft.partial(logpdf, data=data),  # type: ignore
            out_dir=INFERENCE_DIRECTORY,
            labels=labels,
            n_samples=sampler_config["data_dump"]["n_samples"],
        )

        logger.info("Sampling and data saving complete.")


flowMC_arg_parser = guru_arg_parser



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


# Copyright (c) 2024 Colm Talbot
# SPDX-License-Identifier: MIT

import numpyro.distributions
from numpyro.distributions import Distribution


class _Available:
    def __getitem__(self, name: str) -> Distribution:
        if not (ord("A") <= ord(name[0]) <= ord("Z")):
            raise AttributeError(f"module {__name__} is an invalid prior")
        if name in numpyro.distributions.__all__:
            return getattr(numpyro.distributions, name)
        raise AttributeError(f"module {__name__} has no attribute {name}")


available = _Available()



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import functools as ft
from typing import Any, Callable, Dict, Optional, Tuple

import jax
import optax
from jax import numpy as jnp, random as jrd
from jaxtyping import Array, PRNGKeyArray
from numpyro.distributions import Distribution, MultivariateNormal

from gwkokab.models.utils import JointDistribution


@ft.partial(jax.jit, static_argnames=("n_samples",))
def _mvn_samples(loc: Array, cov: Array, n_samples: int, key: PRNGKeyArray) -> Array:
    """Generate samples from a multivariate normal distribution using method from
    `numpyro.distributions.MultivariateNormal`.

    Parameters
    ----------
    loc : Array
        Mean vector of the multivariate normal distribution.
    cov : Array
        Covariance matrix of the multivariate normal distribution.
    n_samples : int
        Number of samples to generate.
    key : PRNGKeyArray
        JAX random key for sampling.

    Returns
    -------
    Array
        Samples drawn from the multivariate normal distribution.
    """
    scale_tril = jnp.linalg.cholesky(cov)
    eps = jrd.normal(key, shape=(n_samples, *loc.shape))
    samples = loc + jnp.squeeze(jnp.matmul(scale_tril, eps[..., jnp.newaxis]), axis=-1)
    return samples


def analytical_likelihood(
    dist_fn: Callable[..., Distribution],
    priors: JointDistribution,
    variables_index: Dict[str, int],
    ERate_fn: Callable[[Distribution], Array],
    key: PRNGKeyArray,
    n_events: int,
    n_samples: int = 10_000,
    max_iter_mean: int = 10,
    max_iter_cov: int = 3,
    n_vi_steps: int = 5,
    learning_rate: float = 1e-2,
    batch_size: int = 1_000,
) -> Callable[[Array, Dict[str, Any]], Array]:
    r"""Compute the analytical likelihood function for a model given its parameters.

    .. math::

        \mathcal{L}_{\mathrm{analytical}}(\Lambda,\kappa)\propto
        \exp\left(-\mu(\Lambda,\kappa)\right)
        \prod_{i=1}^{N}
        \iint\mathcal{G}(\theta,z|\boldsymbol{\mu}_i,\boldsymbol{\Sigma}_i)
        \rho(\theta,z\mid\Lambda,\kappa)d\theta dz

        \mathcal{L}_{\mathrm{analytical}}(\Lambda,\kappa)\propto
        \exp{\left(-\mu(\Lambda,\kappa)\right)}\prod_{i=1}^{N}\bigg<
        \rho(\theta_{i,j},z_{i,j}\mid\Lambda,\kappa)
        \bigg>_{\theta_{i,j},z_{i,j}\sim\mathcal{G}(\theta,z|\boldsymbol{\mu}_i,\boldsymbol{\Sigma}_i)}


    First we are optimizing the mean vector and covariance matrix of the multivariate
    normal distribution that approximates the product of event distributions (also a
    multivariate normal distribution) and phenomenological model distribution by
    Importance Sampling and Moment Matching. After that, we perform Variational
    Inference to optimize the mean vector of the multivariate normal distribution to
    minimize the Reverse KL divergence between the product of event distributions and
    the phenomenological model distribution.

    Parameters
    ----------
    dist_fn : Callable[..., Distribution]
        function that returns a Distribution instance
    priors : JointDistribution
        priors for the model parameters
    variables_index : Dict[str, int]
        mapping of variable names to their indices in the input array
    ERate_fn : Callable[[Distribution], Array]
        function to compute the expected event rates
    means : List[Array]
        List of mean vectors for each event. Each vector should have the same length and
        should correspond to the parameters of the distribution along with covariances.
    covariances : List[Array]
        List of covariance matrices for each event. Each matrix should correspond to the
        mean vectors in `means`.
    key : PRNGKeyArray
        JAX random key for sampling
    n_samples : int, optional
        Number of samples to draw from the multivariate normal distribution for each
        event to compute the likelihood, by default 10_000
    max_iter_mean : int, optional
        Maximum number of iterations for the fitting process of the mean, by default 10
    max_iter_cov : int, optional
        Maximum number of iterations for the fitting process of the covariance, by default 3
    n_vi_steps: int, optional
        Number of steps for the variational inference optimization, by default 5
    learning_rate : float, optional
        Learning rate for the Adam optimizer used in the variational inference
        optimization, by default 1e-2
    batch_size : int, optional
        Batch size for the sampling process, by default 1_000

    Returns
    -------
    Callable[[Array, Array], Array]
        A function that computes the log posterior probability of the model parameters
        given the data. The function takes two arguments: an array of model parameters
        and a second array (not used in this implementation).
    """

    def likelihood_fn(x: Array, data: Dict[str, Any]) -> Array:
        mean_stack: Array = data["mean_stack"]
        cov_stack: Array = data["cov_stack"]
        mapped_params = {
            name: jax.lax.dynamic_index_in_dim(x, i, keepdims=False)
            for name, i in variables_index.items()
        }

        model_instance: Distribution = dist_fn(**mapped_params)

        # μ = E_{Ω|Λ}[VT(ω)]
        expected_rates = ERate_fn(model_instance)

        event_mvn = MultivariateNormal(loc=mean_stack, covariance_matrix=cov_stack)

        rng_key = key
        moment_matching_mean = mean_stack
        moment_matching_cov = event_mvn.covariance_matrix

        model_log_prob_vmap_fn = jax.vmap(
            model_instance.log_prob, in_axes=(1,), out_axes=-1
        )

        normalized_weights_fn = lambda samples: jax.nn.softmax(
            jnp.expand_dims(
                model_log_prob_vmap_fn(samples) + event_mvn.log_prob(samples),
                axis=-1,
            ),
            axis=0,
        )

        @jax.jit
        def scan_moment_matching_mean_fn(
            moment_matching_mean_i: Array, key: PRNGKeyArray
        ) -> Tuple[Array, None]:
            samples = _mvn_samples(
                loc=moment_matching_mean_i,
                cov=moment_matching_cov,
                n_samples=1_000,
                key=key,
            )

            # Normalized weights = softmax(log ρ(samples | Λ, κ) + log G(θ, z | μ_i, Σ_i)))
            weights = normalized_weights_fn(samples)

            # μ_f = sum(weights * samples)
            new_fit_mean = jnp.sum(samples * weights, axis=0)

            return new_fit_mean, None

        rng_key, subkey = jrd.split(rng_key)

        moment_matching_mean, _ = jax.lax.scan(
            scan_moment_matching_mean_fn,  # type: ignore[arg-type]
            moment_matching_mean,
            jrd.split(subkey, max_iter_mean),
            length=max_iter_mean,
        )

        @jax.jit
        def scan_moment_matching_cov_fn(
            moment_matching_cov_i: Array, key: PRNGKeyArray
        ) -> Tuple[Array, None]:
            samples = _mvn_samples(
                loc=moment_matching_mean,
                cov=moment_matching_cov_i,
                n_samples=1_000,
                key=key,
            )

            # Normalized weights = softmax(log ρ(samples | Λ, κ) + log G(θ, z | μ_i, Σ_i)))
            weights = normalized_weights_fn(samples)

            # Σ_f = sum(weights * (samples - μ_f) * (samples - μ_f).T)
            centered = samples - moment_matching_mean
            new_fit_cov = jnp.einsum("sei,sej->eij", weights * centered, centered)

            return new_fit_cov, None

        rng_key, subkey = jrd.split(rng_key)
        moment_matching_cov, _ = jax.lax.scan(
            scan_moment_matching_cov_fn,  # type: ignore[arg-type]
            moment_matching_cov,
            jrd.split(subkey, max_iter_cov),
            length=max_iter_cov,
        )

        @jax.jit
        def loss_fn(mu: Array, cov: Array, key: PRNGKeyArray) -> Array:
            """Compute Reverse KL divergence between the model and the fitted
            multivariate normal distribution.

            Parameters
            ----------
            mu : Array
                mean vector of the multivariate normal distribution
            cov : Array
                covariance matrix of the multivariate normal distribution
            key : PRNGKeyArray
                random key for sampling

            Returns
            -------
            Array
                loss value
            """
            fit_dist = MultivariateNormal(loc=mu, covariance_matrix=cov)
            model_samples = model_instance.sample(key=key, sample_shape=(1_000,))

            log_p = model_instance.log_prob(model_samples)

            return fit_dist.entropy() - jnp.mean(
                jnp.exp(fit_dist.log_prob(model_samples) - log_p) * log_p
            )

        @jax.jit
        def step(
            mu: Array,
            cov: Array,
            opt_state: optax.OptState,
            key: PRNGKeyArray,
        ) -> Tuple[Array, optax.OptState]:
            """Perform a single optimization step to update the mean vector to minimize
            the loss function.

            Parameters
            ----------
            mu : Array
                mean vector of the multivariate normal distribution
            cov : Array
                covariance matrix of the multivariate normal distribution
            opt_state : optax.OptState
                optimizer state
            key : PRNGKeyArray
                random key for sampling

            Returns
            -------
            Tuple[Array, optax.OptState]
                updated mean vector and optimizer state
            """
            keys = jrd.split(key, n_events)
            _, grads = jax.vmap(jax.value_and_grad(loss_fn), in_axes=(0, 0, 0))(
                mu, cov, keys
            )
            updates, opt_state = opt.update(grads, opt_state)
            mu = optax.apply_updates(mu, updates)
            return mu, opt_state

        @jax.jit
        def variational_inference_fn(
            carry: Tuple[Array, Array, optax.OptState, Array], _: Optional[Array]
        ) -> Tuple[Tuple[Array, Array, optax.OptState, Array], None]:
            """Perform a single step of the gradient descent optimization to update the
            mean vector and covariance matrix.

            Parameters
            ----------
            carry : Tuple[Array, Array, optax.OptState, Array]
                carry tuple containing the current mean vector, covariance matrix,
                optimizer state, and random key
            _ : Optional[Array]
                unused placeholder for the scan function

            Returns
            -------
            Tuple[Tuple[Array, Array, optax.OptState, Array], None]
                updated mean vector, unchanged covariance matrix, optimizer state, and
                random key
            """
            mu, cov, opt_state, key = carry
            key, subkey = jrd.split(key)
            mu, opt_state = step(mu, cov, opt_state, subkey)
            return (mu, cov, opt_state, key), None

        vi_mean = moment_matching_mean
        opt = optax.adam(learning_rate=learning_rate)
        opt_state = opt.init(vi_mean)

        rng_key, subkey = jrd.split(rng_key)

        (vi_mean, _, _, _), _ = jax.lax.scan(
            variational_inference_fn,
            (vi_mean, moment_matching_cov, opt_state, subkey),
            length=n_vi_steps,
        )

        fit_mean = jnp.where(jnp.isnan(vi_mean), moment_matching_mean, vi_mean)
        fit_cov = moment_matching_cov

        @jax.jit
        def scan_fn(
            carry: Array, loop_data: Tuple[Array, Array, Array, Array, PRNGKeyArray]
        ) -> Tuple[Array, None]:
            mean, cov, fit_mean_i, fit_cov_i, rng_key_i = loop_data

            # event_mvn = G(θ, z | μ_i, Σ_i)
            event_mvn = MultivariateNormal(loc=mean, covariance_matrix=cov)

            # fit_mvn = G(θ, z | μ_f, Σ_f)
            fit_mvn = MultivariateNormal(loc=fit_mean_i, covariance_matrix=fit_cov_i)

            # data ~ G(θ, z | μ_f, Σ_f)
            data = fit_mvn.sample(rng_key_i, (n_samples,))

            # log ρ(data | Λ, κ)
            model_instance_log_prob = jax.lax.map(
                model_instance.log_prob, data, batch_size=batch_size
            )
            # log G(θ, z | μ_i, Σ_i)
            event_mvn_log_prob = jax.lax.map(
                event_mvn.log_prob, data, batch_size=batch_size
            )
            # log G(θ, z | μ_f, Σ_i)
            fit_mvn_log_prob = jax.lax.map(
                fit_mvn.log_prob, data, batch_size=batch_size
            )

            # log_prob = log ρ(data | Λ, κ) + log G(θ, z | μ_i, Σ_i) - log G(θ, z | μ_f, Σ_i)
            log_prob = model_instance_log_prob + event_mvn_log_prob - fit_mvn_log_prob
            log_likelihood = jax.nn.logsumexp(
                log_prob, axis=-1, where=(~jnp.isneginf(log_prob))
            )
            return carry + log_likelihood, None

        keys = jrd.split(rng_key, (n_events,))

        total_log_likelihood, _ = jax.lax.scan(
            scan_fn,  # type: ignore[arg-type]
            -n_events * jnp.log(n_samples),  # - Σ log(M_i)
            (mean_stack, cov_stack, fit_mean, fit_cov, keys),
            length=n_events,
        )

        # log L(Λ,κ) = -μ + Σ log Σ exp (log ρ(data_n|Λ,κ)) - Σ log(M_i)
        log_likelihood = total_log_likelihood - expected_rates
        # log p(Λ,κ|data) = log π(Λ,κ) + log L(Λ,κ)
        log_posterior = priors.log_prob(x) + log_likelihood

        log_posterior = jnp.nan_to_num(
            log_posterior,
            nan=-jnp.inf,
            posinf=-jnp.inf,
            neginf=-jnp.inf,
        )

        return log_posterior

    return likelihood_fn



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import re
import warnings
from typing import Dict, List


def matches_regex(pattern: str, string: str) -> bool:
    """Check if a string matches a regex pattern.

    Parameters
    ----------
    pattern : str
        regex pattern to match
    string : str
        string to match

    Returns
    -------
    bool
        :code:`True` if the string matches the pattern, `code`:False: otherwise
    """
    return bool(re.fullmatch(pattern, string))


def match_all(
    strings: List[str], pattern_dict_with_val: Dict[str, str | int | float | None]
) -> Dict[str, int | float | None]:
    """Match all strings in a list with a dictionary of regex patterns.

    Parameters
    ----------
    strings : List[str]
        list of strings to match
    pattern_dict_with_val : Dict[str, str  |  int  |  float  |  None]
        dictionary of regex patterns with values

    Returns
    -------
    Dict[str, int | float | None]
        dictionary of matched patterns with values
    """
    # TODO(Qazalbash): Simplify the logic
    matches: Dict[str, int | float | None] = {}
    duplicates = []
    for string in strings:
        if pattern_dict_with_val.get(string):  # Exact match
            matched_string = pattern_dict_with_val[string]
            # Check for duplicates and if the value they are duplicated with is not parsed yet
            # then add them to duplicates list to be parsed later, else add them to matches
            if isinstance(matched_string, str) and matches.get(matched_string) is None:
                duplicates.append(string)
            else:
                matches[string] = matched_string
            continue
        pattern_found = False
        for pattern, value in pattern_dict_with_val.items():
            if matches_regex(pattern, string):
                matches[string] = value
                pattern_found = True
                break
        if not pattern_found:
            matches[string] = None

    for duplicate in duplicates:
        pattern_found = False
        for pattern, value in pattern_dict_with_val.items():
            matched_duplicate = matches_regex(pattern, pattern_dict_with_val[duplicate])
            if not isinstance(value, dict) and matched_duplicate:  # do not match dict
                matches[duplicate] = value
                pattern_found = True
                break
            elif isinstance(value, dict):
                pattern_found = True
                matches[duplicate] = pattern_dict_with_val[duplicate]
        if not pattern_found:
            matches[duplicate] = None

    for string in strings:
        if matches.get(string) is None:
            warnings.warn(f"'{string}' does not match any pattern.")

    return matches



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0
"""Cosmology module for GWKokab, adapted from code written by
`Reed Essick <https://orcid.org/0000-0001-8196-9267>`_ included in the
`gw-distributions <https://git.ligo.org/reed.essick/gw-distributions>`_ package at
`source <https://git.ligo.org/reed.essick/gw-distributions/-/blob/master/gwdistributions/utils/cosmology.py>`_."""

from ._cosmology import Cosmology as Cosmology
from ._planck import (
    PLANCK_2015_Cosmology as PLANCK_2015_Cosmology,
    PLANCK_2018_Cosmology as PLANCK_2018_Cosmology,
)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import argparse
import os
from glob import glob

import arviz as az
import glasbey
import numpy as np
from arviz.utils import _var_names, get_coords
from matplotlib import pyplot as plt


def make_parser() -> argparse.ArgumentParser:
    """Create the command line argument parser.

    This function creates the command line argument parser and returns it.

    Returns
    -------
    argparse.ArgumentParser
        the command line argument parser
    """
    parser = argparse.ArgumentParser(
        description="Pearson correlation coefficient plotter. "
        "source: https://github.com/kazewong/flowMC/blob/main/example/notebook/dualmoon.ipynb (last cell)",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        epilog="This script plots a Pearson correlation coefficients of the chains.",
    )
    parser.add_argument(
        "--chains-regex",
        help="regex pattern for the files containing chains. Only .dat files are supported.",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-o",
        "--output",
        help="output file path",
        required=True,
        type=argparse.FileType("w"),
    )
    parser.add_argument(
        "--n-split",
        help="number of splits",
        default=10,
        type=int,
    )
    parser.add_argument(
        "--title",
        help="title of the plot",
        type=str,
    )
    parser.add_argument(
        "--x-scale",
        help="scale of the x-axis",
        default="linear",
        type=str,
    )
    parser.add_argument(
        "--y-scale",
        help="scale of the y-axis",
        default="linear",
        type=str,
    )
    parser.add_argument(
        "--labels",
        help="labels of the chains",
        nargs="+",
        type=str,
    )
    parser.add_argument(
        "--use-latex",
        help="use LaTeX for rendering text",
        action="store_true",
    )
    parser.add_argument(
        "--font-family",
        help="font family to use",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--dpi",
        help="dots per inch to save file",
        type=int,
        default=100,
    )

    return parser


def main() -> None:
    """Main function of the script."""
    parser = make_parser()
    args = parser.parse_args()

    plt.rcParams.update({"text.usetex": args.use_latex})
    if args.font_family is not None:
        plt.rcParams.update({"font.family": args.font_family})

    chains_filenames = glob(args.chains_regex)
    chains = np.array([np.loadtxt(filename) for filename in chains_filenames])

    n_dims = chains.shape[1]
    plt.rcParams.update(
        {
            "axes.prop_cycle": plt.cycler(
                color=glasbey.create_palette(palette_size=n_dims)
            )
        }
    )

    ## Load data as arviz InferenceData class
    idata = az.convert_to_inference_data(chains)
    coords = {}
    data = get_coords(az.convert_to_dataset(idata, group="posterior"), coords)
    var_names = None
    filter_vars = None
    var_names = _var_names(var_names, data, filter_vars)
    n_draws = data.sizes["draw"]

    first_draw = data.draw.values[0]  # int of where where things should start

    ## Compute where to split the data to diagnostic the convergence
    n_split = args.n_split
    draw_divisions = np.linspace(n_draws // n_split, n_draws, n_split, dtype=int)

    rhat_s = np.stack(
        [
            np.array(
                az.rhat(
                    data.sel(draw=slice(first_draw + draw_div)),
                    var_names=var_names,
                    method="rank",
                )["x"]
            )
            for draw_div in draw_divisions
        ]
    )
    labels = args.labels if args.labels is not None else chains_filenames

    plt.plot(draw_divisions, rhat_s, "-o", label=labels)
    plt.axhline(1, c="k", ls="--")
    plt.xlabel("Iteration")
    plt.ylabel(r"$\hat{R}$")
    plt.legend(
        loc="center left",
        bbox_to_anchor=(1, 0.5),
        fancybox=True,
        shadow=True,
    )
    plt.yscale(args.y_scale)
    plt.xscale(args.x_scale)
    plt.title(args.title)
    plt.tight_layout()
    # Determine output file type and save accordingly
    output_ext = os.path.splitext(args.output.name)[1].lower()
    plt_savefig_kwargs = dict()
    if output_ext == ".png":
        plt_savefig_kwargs["dpi"] = args.dpi
    plt.savefig(args.output.name, bbox_inches="tight", **plt_savefig_kwargs)
    plt.close("all")



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import enum


class Parameters(enum.Enum):
    CHIRP_MASS = "chirp_mass"
    CHIRP_MASS_SOURCE = "chirp_mass_source"
    COS_IOTA = "cos_iota"
    COS_TILT_1 = "cos_tilt_1"
    COS_TILT_2 = "cos_tilt_2"
    DETECTION_TIME = "detection_time"
    ECCENTRICITY = "eccentricity"
    EFFECTIVE_SPIN_MAGNITUDE = "chi_eff"
    MASS_RATIO = "mass_ratio"
    MEAN_ANOMALY = "mean_anomaly"
    PHI_1 = "phi_1"
    PHI_12 = "phi_12"
    PHI_2 = "phi_2"
    PHI_ORB = "phi_orb"
    POLARIZATION_ANGLE = "psi"
    PRIMARY_MASS_DETECTED = "mass_1"
    PRIMARY_MASS_SOURCE = "mass_1_source"
    PRIMARY_SPIN_MAGNITUDE = "a_1"
    PRIMARY_SPIN_X = "spin_1x"
    PRIMARY_SPIN_Y = "spin_1y"
    PRIMARY_SPIN_Z = "spin_1z"
    REDSHIFT = "redshift"
    REDUCED_MASS = "reduced_mass"
    RIGHT_ASCENSION = "ra"
    SECONDARY_MASS_DETECTED = "mass_2"
    SECONDARY_MASS_SOURCE = "mass_2_source"
    SECONDARY_SPIN_MAGNITUDE = "a_2"
    SECONDARY_SPIN_X = "spin_2x"
    SECONDARY_SPIN_Y = "spin_2y"
    SECONDARY_SPIN_Z = "spin_2z"
    SIN_DECLINATION = "dec"
    SYMMETRIC_MASS_RATIO = "symmetric_mass_ratio"



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from . import (
    checks as checks,
    common as common,
    f_monk as f_monk,
    flowMC_based as flowMC_based,
    genie_parser as genie_parser,
    guru as guru,
    poisson_mean_parser as poisson_mean_parser,
    ppd as ppd,
    ppd_parser as ppd_parser,
    regex as regex,
    sage as sage,
)
